from dataclasses import dataclass

@dataclass
class Library:
    ESC: str = ...
    END: str = ...
    FOREGROUND_256: str = ...
    BACKGROUND_256: str = ...
    FOREGROUND_RGB: str = ...
    BACKGROUND_RGB: str = ...
    UNDERLINE_COLOR: str = ...
    CONTROLS = ...
    STYLES = ...
    COLORTERM = ...
    COLORS = ...
    HEX_COLORS = ...
    def __init__(
        self,
        ESC,
        END,
        FOREGROUND_256,
        BACKGROUND_256,
        FOREGROUND_RGB,
        BACKGROUND_RGB,
        UNDERLINE_COLOR,
    ) -> None: ...



import os
import typing
from collections.abc import Mapping

import pytest
from debian.deb822 import Deb822
from debian.debian_support import DpkgArchTable

from debputy._deb_options_profiles import DebBuildOptionsAndProfiles
from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.exceptions import PureVirtualPathError
from debputy.filesystem_scan import OSFSROOverlay
from debputy.highlevel_manifest_parser import YAMLManifestParser
from debputy.manifest_parser.util import AttributePath
from debputy.packages import BinaryPackage, SourcePackage, DctrlParser
from debputy.plugin.api import VirtualPath
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.impl import (
    DebputyPluginInitializerProvider,
    _resolve_bundled_plugin_docs_path,
    find_json_plugin,
)
from debputy.plugin.api.impl_types import (
    DebputyPluginMetadata,
)
from debputy.plugins.debputy.debputy_plugin import initialize_debputy_features
from debputy.substitution import (
    NULL_SUBSTITUTION,
    Substitution,
    SubstitutionImpl,
    VariableContext,
)
from debputy.version import DEBPUTY_PLUGIN_ROOT_DIR

# Disable dpkg's translation layer.  It is very slow and disabling it makes it easier to debug
# test-failure reports from systems with translations active.
os.environ["DPKG_NLS"] = "0"


@pytest.fixture(scope="session")
def amd64_dpkg_architecture_variables() -> DpkgArchitectureBuildProcessValuesTable:
    return DpkgArchitectureBuildProcessValuesTable(fake_host="amd64")


@pytest.fixture(scope="session")
def dpkg_arch_query() -> DpkgArchTable:
    return DpkgArchTable.load_arch_table()


@pytest.fixture()
def source_package() -> SourcePackage:
    return SourcePackage(
        {
            "Source": "foo",
        }
    )


@pytest.fixture()
def package_single_foo_arch_all_cxt_amd64(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
) -> Mapping[str, BinaryPackage]:
    return {
        p.name: p
        for p in [
            BinaryPackage(
                Deb822(
                    {
                        "Package": "foo",
                        "Architecture": "all",
                    }
                ),
                amd64_dpkg_architecture_variables,
                dpkg_arch_query,
                is_main_package=True,
            )
        ]
    }


@pytest.fixture()
def package_foo_w_udeb_arch_any_cxt_amd64(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
) -> Mapping[str, BinaryPackage]:
    return {
        p.name: p
        for p in [
            BinaryPackage(
                Deb822(
                    {
                        "Package": "foo",
                        "Architecture": "any",
                    }
                ),
                amd64_dpkg_architecture_variables,
                dpkg_arch_query,
                is_main_package=True,
            ),
            BinaryPackage(
                Deb822(
                    {
                        "Package": "foo-udeb",
                        "Architecture": "any",
                        "Package-Type": "udeb",
                    }
                ),
                amd64_dpkg_architecture_variables,
                dpkg_arch_query,
            ),
        ]
    }


@pytest.fixture(scope="session")
def null_substitution() -> Substitution:
    return NULL_SUBSTITUTION


@pytest.fixture(scope="function")
def _empty_debputy_plugin_feature_set() -> PluginProvidedFeatureSet:
    return PluginProvidedFeatureSet()


@pytest.fixture(scope="function")
def amd64_substitution(
    amd64_dpkg_architecture_variables,
    _empty_debputy_plugin_feature_set,
) -> Substitution:
    debian_dir = OSFSROOverlay.create_root_dir("debian", "debian")
    variable_context = VariableContext(
        debian_dir,
    )
    return SubstitutionImpl(
        plugin_feature_set=_empty_debputy_plugin_feature_set,
        dpkg_arch_table=amd64_dpkg_architecture_variables,
        static_variables=None,
        environment={},
        unresolvable_substitutions=frozenset(["SOURCE_DATE_EPOCH", "PACKAGE"]),
        variable_context=variable_context,
    )


@pytest.fixture(scope="session")
def no_profiles_or_build_options() -> DebBuildOptionsAndProfiles:
    return DebBuildOptionsAndProfiles(environ={})


@pytest.fixture(scope="function")
def debputy_plugin_feature_set(
    request,
    _empty_debputy_plugin_feature_set,
    amd64_substitution,
) -> PluginProvidedFeatureSet:
    plugin_name = "debputy"
    debputy_plugin_metadata = DebputyPluginMetadata(
        plugin_name=plugin_name,
        api_compat_version=1,
        plugin_initializer=initialize_debputy_features,
        plugin_doc_path_resolver=lambda: _resolve_bundled_plugin_docs_path(
            plugin_name,
            initialize_debputy_features,
        ),
        plugin_loader=None,
        plugin_path="<loaded-via-test>",
    )
    feature_set = _empty_debputy_plugin_feature_set
    api = DebputyPluginInitializerProvider(
        debputy_plugin_metadata,
        feature_set,
        amd64_substitution,
    )
    api.load_plugin()

    plugin_names = getattr(request.function, "_required_plugins", [])
    if plugin_names:
        for plugin_name in plugin_names:
            plugin_metadata = find_json_plugin([DEBPUTY_PLUGIN_ROOT_DIR], plugin_name)
            print(
                f"debputy_plugin_feature_set - Loading extra plugin: {plugin_name} [{plugin_metadata.plugin_path}]"
            )
            plugin_under_test_provider = DebputyPluginInitializerProvider(
                plugin_metadata,
                feature_set,
                amd64_substitution,
            )
            plugin_under_test_provider.load_plugin()

    return feature_set


@pytest.fixture
def attribute_path(request) -> AttributePath:
    return AttributePath.builtin_path()[request.node.nodeid]


class ManifestParserFactory(typing.Protocol):

    def __call__(
        self,
        debian_dir: VirtualPath,
        *,
        dctrl_parser: DctrlParser | None = None,
    ) -> YAMLManifestParser: ...


@pytest.fixture()
def manifest_parser_amd64_factory(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
    amd64_substitution,
    no_profiles_or_build_options,
    debputy_plugin_feature_set,
) -> ManifestParserFactory:

    def _factory(
        debian_dir: VirtualPath,
        *,
        dctrl_parser: DctrlParser | None = None,
    ) -> YAMLManifestParser:
        if dctrl_parser is None:
            dctrl_parser = DctrlParser(
                frozenset(),
                frozenset(),
                False,
                False,
                amd64_dpkg_architecture_variables,
                dpkg_arch_query,
                no_profiles_or_build_options,
            )

        dctrl = debian_dir.get("control")
        try:
            if dctrl is None:
                raise PureVirtualPathError()

            with dctrl.open() as fd:
                _, source_package, binary_packages = (
                    dctrl_parser.parse_source_debian_control(fd.readlines())
                )
        except PureVirtualPathError:
            raise ValueError(
                "The manifest_parser_amd64_factory requires `control` to be present and readable in the `debian_dir`"
            ) from None

        return YAMLManifestParser(
            "debian/test-debputy.manifest",
            source_package,
            binary_packages,
            amd64_substitution,
            # To align with the substitution, we do not allow this field to
            # be overwritten by the DctrlParser
            amd64_dpkg_architecture_variables,
            # Use the same instances
            dctrl_parser.dpkg_arch_query_table,
            dctrl_parser.deb_options_and_profiles,
            debputy_plugin_feature_set,
            "full",
            debian_dir=debian_dir,
        )

    return _factory



import textwrap
from typing import (
    Union,
    Any,
    Optional,
    List,
    Tuple,
    TYPE_CHECKING,
    get_origin,
    Literal,
    get_args,
    Generic,
    cast,
)
from collections.abc import Iterable, Callable, Sequence, Iterator

from debputy.commands.debputy_cmd.output import OutputStyle
from debputy.linting.lint_util import LintState
from debputy.lsp.diagnostics import LintSeverity
from debputy.lsp.quickfixes import propose_correct_text_quick_fix
from debian._deb822_repro.locatable import (
    Position as TEPosition,
    Range as TERange,
)
from debputy.manifest_parser.declarative_parser import (
    DeclarativeMappingInputParser,
    ParserGenerator,
    AttributeDescription,
    DeclarativeNonMappingInputParser,
    BASIC_SIMPLE_TYPES,
)
from debputy.manifest_parser.parser_doc import (
    render_rule,
    render_attribute_doc,
    doc_args_for_parser_doc,
)
from debputy.manifest_parser.tagging_types import DebputyDispatchableType
from debputy.manifest_parser.util import AttributePath
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
from debputy.plugin.api.impl_types import (
    DebputyPluginMetadata,
    DeclarativeInputParser,
    DispatchingParserBase,
    InPackageContextParser,
    ListWrappedDeclarativeInputParser,
    PluginProvidedParser,
    DeclarativeValuelessKeywordInputParser,
    DispatchingTableParser,
)
from debputy.substitution import VariableContext
from debputy.util import _info, _warn, detect_possible_typo, T
from debputy.yaml import MANIFEST_YAML
from debputy.yaml.compat import (
    MarkedYAMLError,
    YAMLError,
)
from debputy.yaml.compat import (
    Node,
    CommentedMap,
    LineCol,
    CommentedSeq,
    CommentedBase,
)

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types

try:
    from pygls.server import LanguageServer
    from debputy.lsp.debputy_ls import DebputyLanguageServer
except ImportError:
    pass


YAML_COMPLETION_HINT_KEY = "___COMPLETE:"
YAML_COMPLETION_HINT_VALUE = "___COMPLETE"
DEBPUTY_PLUGIN_METADATA = plugin_metadata_for_debputys_own_plugin()


class LSPYAMLHelper(Generic[T]):

    def __init__(
        self,
        lint_state: LintState,
        pg: ParserGenerator,
        custom_data: T,
    ) -> None:
        self.lint_state = lint_state
        self.lines = _lines(lint_state.lines)
        self.pg = pg
        self.custom_data = custom_data

    def _validate_subparser_is_valid_here(
        self,
        subparser: PluginProvidedParser,
        orig_key: str,
        line: int,
        col: int,
    ) -> None:
        # Subclasses can provide custom logic here
        pass

    def _lint_dispatch_parser(
        self,
        parser: DispatchingParserBase,
        dispatch_key: str,
        key_pos: tuple[int, int] | None,
        value: Any | None,
        value_pos: tuple[int, int] | None,
        *,
        is_keyword_only: bool,
    ) -> None:
        is_known = parser.is_known_keyword(dispatch_key)
        orig_key = dispatch_key
        if not is_known and key_pos is not None:

            if value is None:
                opts = {
                    "message_format": 'Unknown or unsupported value "{key}".',
                }
            else:
                opts = {}
            corrected_key = yaml_flag_unknown_key(
                self.lint_state,
                dispatch_key,
                parser.registered_keywords(),
                key_pos,
                unknown_keys_diagnostic_severity=parser.unknown_keys_diagnostic_severity,
                **opts,
            )
            if corrected_key is not None:
                dispatch_key = corrected_key
                is_known = True

        if is_known:
            subparser = parser.parser_for(dispatch_key)
            assert subparser is not None
            if key_pos:
                line, col = key_pos
                self._validate_subparser_is_valid_here(
                    subparser,
                    orig_key,
                    line,
                    col,
                )

            if isinstance(subparser.parser, DeclarativeValuelessKeywordInputParser):
                if value is not None or not is_keyword_only:
                    if value is not None:
                        line_no, cursor_pos = value_pos if value_pos else key_pos
                        value_range = self._remaining_line(line_no, cursor_pos)
                        if _is_empty_range(value_range):
                            # In the unlikely case that the value position is present but leads to
                            # an empty range, report the key instead.
                            line_no, cursor_pos = key_pos
                            value_range = self._remaining_line(line_no, cursor_pos)
                        msg = f"The keyword {dispatch_key} does not accept any value"
                    else:
                        line_no, cursor_pos = key_pos
                        value_range = self._remaining_line(line_no, cursor_pos)
                        msg = f"The keyword {dispatch_key} cannot be used as a mapping key"

                    assert not _is_empty_range(value_range)
                    self.lint_state.emit_diagnostic(
                        value_range,
                        msg,
                        "error",
                        "debputy",
                    )
                return

            self.lint_content(
                # Pycharm's type checking gets confused by the isinstance check above.
                cast("DeclarativeInputParser[Any]", subparser.parser),
                value,
                key=orig_key,
                content_pos=value_pos,
            )

    def lint_content(
        self,
        parser: DeclarativeInputParser[Any],
        content: Any,
        *,
        key: str | int | None = None,
        content_pos: tuple[int, int] | None = None,
    ) -> None:
        if isinstance(parser, DispatchingParserBase):
            if isinstance(content, str):
                self._lint_dispatch_parser(
                    parser,
                    content,
                    content_pos,
                    None,
                    None,
                    is_keyword_only=True,
                )

                return
            if not isinstance(content, CommentedMap):
                return
            lc = content.lc
            for dispatch_key, value in content.items():
                key_pos = lc.key(dispatch_key)
                value_pos = lc.value(dispatch_key)
                self._lint_dispatch_parser(
                    parser,
                    dispatch_key,
                    key_pos,
                    value,
                    value_pos,
                    is_keyword_only=False,
                )

        elif isinstance(parser, ListWrappedDeclarativeInputParser):
            if not isinstance(content, CommentedSeq):
                return
            subparser = parser.delegate
            lc = content.lc
            for idx, value in enumerate(content):
                value_pos = lc.item(idx)
                self.lint_content(subparser, value, content_pos=value_pos, key=idx)
        elif isinstance(parser, InPackageContextParser):
            if not isinstance(content, CommentedMap):
                return
            known_packages = self.lint_state.binary_packages
            lc = content.lc
            for k, v in content.items():
                if k is None or (
                    "{{" not in k
                    and known_packages is not None
                    and k not in known_packages
                ):
                    yaml_flag_unknown_key(
                        self.lint_state,
                        k,
                        known_packages,
                        lc.key(k),
                        message_format='Unknown package "{key}".',
                    )
                self.lint_content(parser.delegate, v, key=k, content_pos=lc.value(k))
        elif isinstance(parser, DeclarativeMappingInputParser):
            self._lint_declarative_mapping_input_parser(
                parser,
                content,
                content_pos,
                key=key,
            )
        elif isinstance(parser, DeclarativeNonMappingInputParser):
            if content_pos is not None:
                self._lint_attr_value(
                    parser.alt_form_parser,
                    key,
                    content,
                    content_pos,
                )

    def _lint_declarative_mapping_input_parser(
        self,
        parser: DeclarativeMappingInputParser,
        content: Any,
        content_pos: tuple[int, int],
        *,
        key: str | int | None = None,
    ) -> None:
        if not isinstance(content, CommentedMap):
            alt_form_parser = parser.alt_form_parser
            if alt_form_parser:
                self._lint_attr_value(
                    alt_form_parser,
                    key,
                    content,
                    content_pos,
                )
            else:
                line_no, cursor_pos = content_pos
                value_range = self._remaining_line(line_no, cursor_pos)
                if _is_empty_range(value_range):
                    # FIXME: We cannot report an empty range, but there is still a problem here.
                    return
                if isinstance(key, str):
                    msg = f'The value for "{key}" must be a mapping'
                else:
                    msg = "The value must be a mapping"
                self.lint_state.emit_diagnostic(
                    value_range,
                    msg,
                    "error",
                    "debputy",
                )
            return
        lc = content.lc
        for key, value in content.items():
            attr = parser.manifest_attributes.get(key)
            key_pos = lc.key(key)
            value_pos = lc.value(key)
            if attr is None:
                corrected_key = yaml_flag_unknown_key(
                    self.lint_state,
                    key,
                    parser.manifest_attributes,
                    key_pos,
                )
                if corrected_key:
                    key = corrected_key
                    attr = parser.manifest_attributes.get(corrected_key)
            if attr is None:
                continue

            self._lint_attr_value(
                attr,
                key,
                value,
                value_pos,
            )

            for forbidden_key in attr.conflicting_attributes:
                if forbidden_key in content:
                    line, col = key_pos
                    con_line, con_col = lc.key(forbidden_key)
                    yaml_conflicting_key(
                        self.lint_state,
                        key,
                        forbidden_key,
                        line,
                        col,
                        con_line,
                        con_col,
                    )
        for mx in parser.mutually_exclusive_attributes:
            matches = content.keys() & mx
            if len(matches) < 2:
                continue
            key, *others = list(matches)
            line, col = lc.key(key)
            for other in others:
                con_line, con_col = lc.key(other)
                yaml_conflicting_key(
                    self.lint_state,
                    key,
                    other,
                    line,
                    col,
                    con_line,
                    con_col,
                )

    def _type_based_value_check(
        self,
        target_attr_type: type,
        value: Any,
        value_pos: tuple[int, int],
        *,
        key: str | int | None = None,
    ) -> bool:
        if issubclass(target_attr_type, DebputyDispatchableType):
            parser = self.pg.dispatch_parser_table_for(target_attr_type)
            self.lint_content(
                parser,
                value,
                key=key,
                content_pos=value_pos,
            )
            return True
        return False

    def _lint_attr_value(
        self,
        attr: AttributeDescription,
        key: str | int | None,
        value: Any,
        pos: tuple[int, int],
    ) -> None:
        target_attr_type = attr.attribute_type
        orig = get_origin(target_attr_type)
        if orig == list and isinstance(value, CommentedSeq):
            lc = value.lc
            target_item_type = get_args(target_attr_type)[0]
            for idx, v in enumerate(value):
                v_pos = lc.item(idx)
                self._lint_value(
                    idx,
                    v,
                    target_item_type,
                    v_pos,
                )

        else:
            self._lint_value(
                key,
                value,
                target_attr_type,
                pos,
            )

    def _lint_value(
        self,
        key: str | int | None,
        value: Any,
        target_attr_type: Any,
        pos: tuple[int, int],
    ) -> None:
        type_mapping = self.pg.get_mapped_type_from_target_type(target_attr_type)
        source_attr_type = target_attr_type
        if type_mapping is not None:
            source_attr_type = type_mapping.source_type
        valid_values: Sequence[Any] | None = None
        orig = get_origin(source_attr_type)
        if orig == Literal:
            valid_values = get_args(target_attr_type)
        elif orig == bool or target_attr_type == bool:
            valid_values = (True, False)
        elif isinstance(target_attr_type, type) and self._type_based_value_check(
            target_attr_type,
            value,
            pos,
            key=key,
        ):
            return
        elif source_attr_type in BASIC_SIMPLE_TYPES:
            if isinstance(value, source_attr_type):
                return
            expected_type = BASIC_SIMPLE_TYPES[source_attr_type]
            line_no, cursor_pos = pos
            value_range = self._remaining_line(line_no, cursor_pos)
            if _is_empty_range(value_range):
                # FIXME: We cannot report an empty range, but there is still a problem here.
                return
            if isinstance(key, str):
                msg = f'Value for "{key}" does not match the base type: Expected {expected_type}'
            else:
                msg = f"Value does not match the base type: Expected {expected_type}"
            if issubclass(source_attr_type, str):
                quickfixes = [
                    propose_correct_text_quick_fix(_as_yaml_value(str(value)))
                ]
            else:
                quickfixes = None
            self.lint_state.emit_diagnostic(
                value_range,
                msg,
                "error",
                "debputy",
                quickfixes=quickfixes,
            )
            return

        if valid_values is None or value in valid_values:
            return
        line_no, cursor_pos = pos
        value_range = self._remaining_line(line_no, cursor_pos)
        if _is_empty_range(value_range):
            # FIXME: We cannot report an empty range, but there is still a problem here.
            return
        if isinstance(key, str):
            msg = f'Not a supported value for "{key}"'
        else:
            msg = "Not a supported value here"
        self.lint_state.emit_diagnostic(
            value_range,
            msg,
            "error",
            "debputy",
            quickfixes=[
                propose_correct_text_quick_fix(_as_yaml_value(m)) for m in valid_values
            ],
        )

    def _remaining_line(self, line_no: int, pos_start: int) -> "TERange":
        raw_line = self.lines[line_no].rstrip()
        pos_end = len(raw_line)
        return TERange(
            TEPosition(
                line_no,
                pos_start,
            ),
            TEPosition(
                line_no,
                pos_end,
            ),
        )


def _is_empty_range(token_range: "TERange") -> bool:
    return token_range.start_pos == token_range.end_pos


def _lines(lines: list[str]) -> list[str]:
    if not lines or lines[-1].endswith("\n"):
        lines = lines.copy()
        lines.append("")
    return lines


async def generic_yaml_lint(
    lint_state: LintState,
    root_parser: DeclarativeInputParser[Any],
    initialize_yaml_helper: Callable[[LintState], LSPYAMLHelper[Any]],
) -> None:
    lines = _lines(lint_state.lines)
    try:
        content = MANIFEST_YAML.load(lint_state.content)
    except MarkedYAMLError as e:
        if e.context_mark:
            line = e.context_mark.line
            column = e.context_mark.column
        else:
            line = e.problem_mark.line
            column = e.problem_mark.column
        error_range = error_range_at_position(
            lines,
            line,
            column,
        )
        lint_state.emit_diagnostic(
            error_range,
            f"YAML parse error: {e}",
            "error",
            "debputy",
        )
    except YAMLError as e:
        error_range = TERange(
            TEPosition(0, 0),
            TEPosition(0, len(lines[0])),
        )
        lint_state.emit_diagnostic(
            error_range,
            f"Unknown YAML parse error: {e} [{e!r}]",
            "error",
            "debputy",
        )
    else:
        yaml_linter = initialize_yaml_helper(lint_state)
        yaml_linter.lint_content(
            root_parser,
            content,
        )


def _as_yaml_value(v: Any) -> str:
    if isinstance(v, bool):
        return str(v).lower()
    if isinstance(v, str):
        return maybe_quote_yaml_value(str(v))
    return str(v)


def resolve_hover_text_for_value(
    feature_set: PluginProvidedFeatureSet,
    parser: DeclarativeMappingInputParser,
    plugin_metadata: DebputyPluginMetadata,
    output_style: OutputStyle,
    show_integration_mode: bool,
    segment: str | int,
    matched: Any,
) -> str | None:

    hover_doc_text: str | None = None
    attr = parser.manifest_attributes.get(segment)
    attr_type = attr.attribute_type if attr is not None else None
    if attr_type is None:
        _info(f"Matched value for {segment} -- No attr or type")
        return None
    if isinstance(attr_type, type) and issubclass(attr_type, DebputyDispatchableType):
        parser_generator = feature_set.manifest_parser_generator
        parser = parser_generator.dispatch_parser_table_for(attr_type)
        if parser is None or not isinstance(matched, str):
            _info(
                f"Unknown parser for {segment} or matched is not a str -- {attr_type} {type(matched)=}"
            )
            return None
        subparser = parser.parser_for(matched)
        if subparser is None:
            _info(f"Unknown parser for {matched} (subparser)")
            return None
        hover_doc_text = render_rule(
            matched,
            subparser.parser,
            plugin_metadata,
            output_style,
            show_integration_mode=show_integration_mode,
        )
    else:
        _info(f"Unknown value: {matched} -- {segment}")
    return hover_doc_text


def resolve_hover_text(
    feature_set: PluginProvidedFeatureSet,
    parser: DeclarativeInputParser[Any] | DispatchingParserBase | None,
    plugin_metadata: DebputyPluginMetadata,
    output_style: OutputStyle,
    show_integration_mode: bool,
    segments: list[str | int],
    at_depth_idx: int,
    matched: Any,
    matched_key: bool,
) -> str | None:
    hover_doc_text: str | None = None
    if at_depth_idx == len(segments):
        segment = segments[at_depth_idx - 1]
        _info(f"Matched {segment} at ==, {matched_key=} ")
        hover_doc_text = render_rule(
            segment,
            parser,
            plugin_metadata,
            output_style,
            is_root_rule=False,
            show_integration_mode=show_integration_mode,
        )
    elif at_depth_idx + 1 == len(segments) and isinstance(
        parser, DeclarativeMappingInputParser
    ):
        segment = segments[at_depth_idx]
        _info(f"Matched {segment} at -1, {matched_key=} ")
        if isinstance(segment, str):
            if not matched_key:
                hover_doc_text = resolve_hover_text_for_value(
                    feature_set,
                    parser,
                    plugin_metadata,
                    output_style,
                    show_integration_mode,
                    segment,
                    matched,
                )
            if matched_key or hover_doc_text is None:
                rule_name = _guess_rule_name(segments, at_depth_idx)
                hover_doc_text = _render_param_doc(
                    rule_name,
                    parser,
                    plugin_metadata,
                    segment,
                )
    else:
        _info(f"No doc: {at_depth_idx=} {len(segments)=}")

    return hover_doc_text


def as_hover_doc(
    ls: "DebputyLanguageServer",
    hover_doc_text: str | None,
) -> types.Hover | None:
    if hover_doc_text is None:
        return None
    return types.Hover(
        contents=types.MarkupContent(
            kind=ls.hover_markup_format(
                types.MarkupKind.Markdown,
                types.MarkupKind.PlainText,
            ),
            value=hover_doc_text,
        ),
    )


def _render_param_doc(
    rule_name: str,
    declarative_parser: DeclarativeMappingInputParser,
    plugin_metadata: DebputyPluginMetadata,
    attribute: str,
) -> str | None:
    attr = declarative_parser.source_attributes.get(attribute)
    if attr is None:
        return None

    doc_args, parser_doc = doc_args_for_parser_doc(
        rule_name,
        declarative_parser,
        plugin_metadata,
    )
    rendered_docs = render_attribute_doc(
        declarative_parser,
        declarative_parser.source_attributes,
        declarative_parser.input_time_required_parameters,
        declarative_parser.at_least_one_of,
        parser_doc,
        doc_args,
        is_interactive=True,
        rule_name=rule_name,
    )

    for attributes, rendered_doc in rendered_docs:
        if attribute in attributes:
            full_doc = [
                f"# Attribute `{attribute}`",
                "",
            ]
            full_doc.extend(rendered_doc)

            return "\n".join(full_doc)
    return None


def _guess_rule_name(segments: list[str | int], idx: int) -> str:
    orig_idx = idx
    idx -= 1
    while idx >= 0:
        segment = segments[idx]
        if isinstance(segment, str):
            return segment
        idx -= 1
    _warn(f"Unable to derive rule name from {segments} [{orig_idx}]")
    return "<Bug: unknown rule name>"


def is_at(position: types.Position, lc_pos: tuple[int, int]) -> bool:
    return position.line == lc_pos[0] and position.character == lc_pos[1]


def is_before(position: types.Position, lc_pos: tuple[int, int]) -> bool:
    line, column = lc_pos
    if position.line < line:
        return True
    if position.line == line and position.character < column:
        return True
    return False


def is_after(position: types.Position, lc_pos: tuple[int, int]) -> bool:
    line, column = lc_pos
    if position.line > line:
        return True
    if position.line == line and position.character > column:
        return True
    return False


def error_range_at_position(
    lines: list[str],
    line_no: int,
    char_offset: int,
) -> TERange:
    line = lines[line_no]
    line_len = len(line)
    start_idx = char_offset
    end_idx = start_idx

    if line[start_idx].isspace():

        def _check(x: str) -> bool:
            return not x.isspace()

    else:

        def _check(x: str) -> bool:
            return x.isspace()

    for i in range(end_idx, line_len):
        end_idx = i
        if _check(line[i]):
            break

    for i in range(start_idx, -1, -1):
        if i > 0 and _check(line[i]):
            break
        start_idx = i

    return TERange(
        TEPosition(line_no, start_idx),
        TEPosition(line_no, end_idx),
    )


def _escape(v: str) -> str:
    return '"' + v.replace("\n", "\\n") + '"'


def insert_complete_marker_snippet(
    lines: list[str],
    server_position: types.Position,
) -> bool:
    _info(f"Complete at {server_position}")
    line_no = server_position.line
    line = lines[line_no] if line_no < len(lines) else ""

    lhs_ws = line[: server_position.character]
    lhs = lhs_ws.strip()
    open_quote = ""
    rhs = line[server_position.character + 1 :]
    for q in ('"', "'"):
        if rhs.endswith(q):
            break
        qc = lhs.count(q) & 1
        if qc:
            open_quote = q
            break

    if lhs.endswith(":"):
        _info("Insertion of value (key seen)")
        new_line = (
            line[: server_position.character]
            + YAML_COMPLETION_HINT_VALUE
            + f"{open_quote}\n"
        )
    elif lhs.startswith("-"):
        _info("Insertion of key or value (list item)")
        # Respect the provided indentation
        snippet = (
            YAML_COMPLETION_HINT_KEY if ":" not in lhs else YAML_COMPLETION_HINT_VALUE
        )
        new_line = line[: server_position.character] + snippet + f"{open_quote}\n"
    elif not lhs or (lhs_ws and not lhs_ws[0].isspace()):
        _info(f"Insertion of key or value: {_escape(line[server_position.character:])}")
        # Respect the provided indentation
        snippet = (
            YAML_COMPLETION_HINT_KEY if ":" not in lhs else YAML_COMPLETION_HINT_VALUE
        )
        new_line = line[: server_position.character] + snippet + f"{open_quote}\n"
    elif lhs.isalpha() and ":" not in lhs:
        _info(f"Expanding value to a key: {_escape(line[server_position.character:])}")
        # Respect the provided indentation
        new_line = (
            line[: server_position.character]
            + YAML_COMPLETION_HINT_KEY
            + f"{open_quote}\n"
        )
    elif open_quote:
        _info(
            f"Expanding value inside a string: {_escape(line[server_position.character:])}"
        )
        new_line = (
            line[: server_position.character]
            + YAML_COMPLETION_HINT_VALUE
            + f"{open_quote}\n"
        )
    else:
        c = (
            line[server_position.character]
            if server_position.character < len(line)
            else "(OOB)"
        )
        _info(f"Not touching line: {_escape(line)} -- {_escape(c)}")
        return False
    _info(f'Evaluating complete on synthetic line: "{new_line}"')
    if line_no < len(lines):
        lines[line_no] = new_line
    elif line_no == len(lines):
        lines.append(new_line)
    else:
        return False
    return True


def _keywords_with_parser(
    parser: DeclarativeMappingInputParser | DispatchingParserBase,
) -> Iterator[tuple[str, PluginProvidedParser]]:
    for keyword in parser.registered_keywords():
        pp_subparser = parser.parser_for(keyword)
        yield keyword, pp_subparser


def yaml_key_range(
    key: str | None,
    line: int,
    col: int,
) -> "TERange":
    key_len = len(key) if key else 1
    return TERange.between(
        TEPosition(line, col),
        TEPosition(line, col + key_len),
    )


def yaml_flag_unknown_key(
    lint_state: LintState,
    key: str | None,
    expected_keys: Iterable[str],
    key_pos: tuple[int, int],
    *,
    message_format: str = 'Unknown or unsupported key "{key}".',
    unknown_keys_diagnostic_severity: LintSeverity | None = "error",
) -> str | None:
    line, col = key_pos
    key_range = yaml_key_range(key, line, col)

    candidates = detect_possible_typo(key, expected_keys) if key is not None else ()
    extra = ""
    corrected_key = None
    if candidates:
        extra = f' It looks like a typo of "{candidates[0]}".'
        # TODO: We should be able to tell that `install-doc` and `install-docs` are the same.
        #  That would enable this to work in more cases.
        corrected_key = candidates[0] if len(candidates) == 1 else None
        if unknown_keys_diagnostic_severity is None:
            message_format = f"Possible typo of {candidates[0]}."
            extra = ""
    elif unknown_keys_diagnostic_severity is None:
        return None

    if key is None:
        message_format = "Missing key"
    if unknown_keys_diagnostic_severity is not None:
        lint_state.emit_diagnostic(
            key_range,
            message_format.format(key=key) + extra,
            unknown_keys_diagnostic_severity,
            "debputy",
            quickfixes=[propose_correct_text_quick_fix(n) for n in candidates],
        )
    return corrected_key


def yaml_conflicting_key(
    lint_state: LintState,
    key_a: str,
    key_b: str,
    key_a_line: int,
    key_a_col: int,
    key_b_line: int,
    key_b_col: int,
) -> None:
    key_a_range = TERange(
        TEPosition(
            key_a_line,
            key_a_col,
        ),
        TEPosition(
            key_a_line,
            key_a_col + len(key_a),
        ),
    )
    key_b_range = TERange(
        TEPosition(
            key_b_line,
            key_b_col,
        ),
        TEPosition(
            key_b_line,
            key_b_col + len(key_b),
        ),
    )
    lint_state.emit_diagnostic(
        key_a_range,
        f'The "{key_a}" cannot be used with "{key_b}".',
        "error",
        "debputy",
        related_information=[
            lint_state.related_diagnostic_information(
                key_b_range, f'The attribute "{key_b}" is used here.'
            ),
        ],
    )

    lint_state.emit_diagnostic(
        key_b_range,
        f'The "{key_b}" cannot be used with "{key_a}".',
        "error",
        "debputy",
        related_information=[
            lint_state.related_diagnostic_information(
                key_a_range,
                f'The attribute "{key_a}" is used here.',
            ),
        ],
    )


def resolve_keyword(
    current_parser: DeclarativeInputParser[Any] | DispatchingParserBase,
    current_plugin: DebputyPluginMetadata,
    segments: list[str | int],
    segment_idx: int,
    parser_generator: ParserGenerator,
    *,
    is_completion_attempt: bool = False,
) -> None | (
    tuple[
        DeclarativeInputParser[Any] | DispatchingParserBase,
        DebputyPluginMetadata,
        int,
    ]
):
    if segment_idx >= len(segments):
        return current_parser, current_plugin, segment_idx
    current_segment = segments[segment_idx]
    if isinstance(current_parser, ListWrappedDeclarativeInputParser):
        if isinstance(current_segment, int):
            current_parser = current_parser.delegate
            segment_idx += 1
            if segment_idx >= len(segments):
                return current_parser, current_plugin, segment_idx
            current_segment = segments[segment_idx]

    if not isinstance(current_segment, str):
        return None

    if is_completion_attempt and current_segment.endswith(
        (YAML_COMPLETION_HINT_KEY, YAML_COMPLETION_HINT_VALUE)
    ):
        return current_parser, current_plugin, segment_idx

    if isinstance(current_parser, InPackageContextParser):
        return resolve_keyword(
            current_parser.delegate,
            current_plugin,
            segments,
            segment_idx + 1,
            parser_generator,
            is_completion_attempt=is_completion_attempt,
        )
    elif isinstance(current_parser, DispatchingParserBase):
        if not current_parser.is_known_keyword(current_segment):
            if is_completion_attempt:
                return current_parser, current_plugin, segment_idx
            return None
        subparser = current_parser.parser_for(current_segment)
        segment_idx += 1
        if segment_idx < len(segments):
            return resolve_keyword(
                subparser.parser,
                subparser.plugin_metadata,
                segments,
                segment_idx,
                parser_generator,
                is_completion_attempt=is_completion_attempt,
            )
        return subparser.parser, subparser.plugin_metadata, segment_idx
    elif isinstance(current_parser, DeclarativeMappingInputParser):
        attr = current_parser.manifest_attributes.get(current_segment)
        attr_type = attr.attribute_type if attr is not None else None
        if (
            attr_type is not None
            and isinstance(attr_type, type)
            and issubclass(attr_type, DebputyDispatchableType)
        ):
            subparser = parser_generator.dispatch_parser_table_for(attr_type)
            if subparser is not None and (
                is_completion_attempt or segment_idx + 1 < len(segments)
            ):
                return resolve_keyword(
                    subparser,
                    current_plugin,
                    segments,
                    segment_idx + 1,
                    parser_generator,
                    is_completion_attempt=is_completion_attempt,
                )
        return current_parser, current_plugin, segment_idx
    else:
        _info(f"Unknown parser: {current_parser.__class__}")
    return None


def _trace_cursor(
    content: Any,
    attribute_path: AttributePath,
    server_position: types.Position,
) -> tuple[bool, AttributePath, Any, Any] | None:
    matched_key: str | int | None = None
    matched: Node | None = None
    matched_was_key: bool = False

    if isinstance(content, CommentedMap):
        dict_lc: LineCol = content.lc
        for k, v in content.items():
            k_lc = dict_lc.key(k)
            if is_before(server_position, k_lc):
                break
            v_lc = dict_lc.value(k)
            if is_before(server_position, v_lc):
                # TODO: Handle ":" and "whitespace"
                matched = k
                matched_key = k
                matched_was_key = True
                break
            matched = v
            matched_key = k
    elif isinstance(content, CommentedSeq):
        list_lc: LineCol = content.lc
        for idx, value in enumerate(content):
            i_lc = list_lc.item(idx)
            if is_before(server_position, i_lc):
                break
            matched_key = idx
            matched = value

    if matched is not None:
        assert matched_key is not None
        sub_path = attribute_path[matched_key]
        if not matched_was_key and isinstance(matched, CommentedBase):
            return _trace_cursor(matched, sub_path, server_position)
        return matched_was_key, sub_path, matched, content
    return None


def maybe_quote_yaml_value(v: str) -> str:
    if v and v[0].isdigit():
        try:
            float(v)
            return f"'{v}'"
        except ValueError:
            pass
    return v


def _complete_value(v: Any) -> str:
    if isinstance(v, str):
        return maybe_quote_yaml_value(v)
    return str(v)


def completion_from_attr(
    attr: AttributeDescription,
    pg: ParserGenerator,
    matched: Any,
    *,
    matched_key: bool = False,
    has_colon: bool = False,
) -> types.CompletionList | Sequence[types.CompletionItem] | None:
    type_mapping = pg.get_mapped_type_from_target_type(attr.attribute_type)
    if type_mapping is not None:
        attr_type = type_mapping.source_type
    else:
        attr_type = attr.attribute_type

    orig = get_origin(attr_type)
    valid_values: Sequence[Any] = tuple()

    if orig == Literal:
        valid_values = get_args(attr_type)
    elif orig == bool or attr.attribute_type == bool:
        valid_values = ("true", "false")
    elif isinstance(attr_type, type) and issubclass(attr_type, DebputyDispatchableType):
        parser: DispatchingTableParser[Any] = pg.dispatch_parser_table_for(attr_type)
        if parser is None:
            return None
        valid_values = [
            k if has_colon or not matched_key else f"{k}:"
            for k in parser.registered_keywords()
            if isinstance(
                parser.parser_for(k).parser, DeclarativeValuelessKeywordInputParser
            )
            ^ matched_key
        ]

    if matched in valid_values:
        _info(f"Already filled: {matched} is one of {valid_values}")
        return None
    if valid_values:
        return [types.CompletionItem(_complete_value(x)) for x in valid_values]
    return None


def completion_item(
    quoted_keyword: str,
    pp_subparser: PluginProvidedParser,
) -> types.CompletionItem:
    inline_reference_documentation = pp_subparser.parser.inline_reference_documentation
    synopsis = (
        inline_reference_documentation.synopsis
        if inline_reference_documentation
        else None
    )
    return types.CompletionItem(
        quoted_keyword,
        detail=synopsis,
    )


def _is_inside_manifest_variable_substitution(
    lines: list[str],
    server_position: types.Position,
) -> bool:

    current_line = lines[server_position.line]
    try:
        open_idx = current_line[0 : server_position.character].rindex("{{")
        return "}}" not in current_line[open_idx : server_position.character]
    except ValueError:
        return False


def _manifest_substitution_variable_at_position(
    lines: list[str],
    server_position: types.Position,
) -> str | None:
    current_line = lines[server_position.line]
    try:
        open_idx = current_line[0 : server_position.character].rindex("{{") + 2
        if "}}" in current_line[open_idx : server_position.character]:
            return None
        variable_len = current_line[open_idx:].index("}}")
        close_idx = open_idx + variable_len
    except ValueError as e:
        return None
    return current_line[open_idx:close_idx]


def _insert_complete_marker_and_parse_yaml(
    lines: list[str],
    server_position: types.Position,
) -> Any | None:
    added_key = insert_complete_marker_snippet(lines, server_position)
    attempts = 1 if added_key else 2
    content = None
    while attempts > 0:
        attempts -= 1
        try:
            # Since we mutated the lines to insert a token, `doc.source` cannot
            # be used here.
            content = MANIFEST_YAML.load("".join(lines))
            break
        except MarkedYAMLError as e:
            context_line = (
                e.context_mark.line if e.context_mark else e.problem_mark.line
            )
            if (
                e.problem_mark.line != server_position.line
                and context_line != server_position.line
            ):
                l_data = (
                    lines[e.problem_mark.line].rstrip()
                    if e.problem_mark.line < len(lines)
                    else "N/A (OOB)"
                )

                _info(f"Parse error on line: {e.problem_mark.line}: {l_data}")
                return None

            if attempts > 0:
                # Try to make it a key and see if that fixes the problem
                new_line = (
                    lines[server_position.line].rstrip() + YAML_COMPLETION_HINT_KEY
                )
                lines[server_position.line] = new_line
        except YAMLError:
            break
    return content


def generic_yaml_completer(
    ls: "DebputyLanguageServer",
    params: types.CompletionParams,
    root_parser: DeclarativeInputParser[Any],
) -> types.CompletionList | Sequence[types.CompletionItem] | None:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lines = _lines(doc.lines)
    server_position = doc.position_codec.position_from_client_units(
        lines, params.position
    )
    orig_line = lines[server_position.line].rstrip()
    has_colon = ":" in orig_line

    content = _insert_complete_marker_and_parse_yaml(lines, server_position)
    if content is None:
        context = lines[server_position.line].replace("\n", "\\n")
        _info(f"Completion failed: parse error: Line in question: {context}")
        return None
    attribute_root_path = AttributePath.root_path(content)
    m = _trace_cursor(content, attribute_root_path, server_position)

    if m is None:
        _info("No match")
        return None
    matched_key, attr_path, matched, parent = m
    _info(f"Matched path: {matched} (path: {attr_path.path}) [{matched_key=}]")
    feature_set = ls.plugin_feature_set
    segments = list(attr_path.path_segments())
    km = resolve_keyword(
        root_parser,
        DEBPUTY_PLUGIN_METADATA,
        segments,
        0,
        feature_set.manifest_parser_generator,
        is_completion_attempt=True,
    )
    if km is None:
        return None
    parser, _, at_depth_idx = km
    _info(f"Match leaf parser {at_depth_idx} -- {parser.__class__}")
    items = []
    if at_depth_idx + 1 < len(segments):
        return items

    if _is_inside_manifest_variable_substitution(lines, server_position):
        return [
            types.CompletionItem(
                pv.variable_name,
                detail=pv.variable_reference_documentation,
                sort_text=(
                    f"zz-{pv.variable_name}"
                    if pv.is_for_special_case
                    else pv.variable_name
                ),
            )
            for pv in ls.plugin_feature_set.manifest_variables.values()
            if not pv.is_internal
        ]

    if isinstance(parser, DispatchingParserBase):
        if matched_key:
            items = [
                completion_item(
                    (
                        maybe_quote_yaml_value(k)
                        if has_colon
                        else f"{maybe_quote_yaml_value(k)}:"
                    ),
                    pp_subparser,
                )
                for k, pp_subparser in _keywords_with_parser(parser)
                if k not in parent
                and not isinstance(
                    pp_subparser.parser,
                    DeclarativeValuelessKeywordInputParser,
                )
            ]
        else:
            items = [
                completion_item(maybe_quote_yaml_value(k), pp_subparser)
                for k, pp_subparser in _keywords_with_parser(parser)
                if k not in parent
                and isinstance(
                    pp_subparser.parser,
                    DeclarativeValuelessKeywordInputParser,
                )
            ]
    elif isinstance(parser, InPackageContextParser):
        binary_packages = ls.lint_state(doc).binary_packages
        if binary_packages is not None:
            items = [
                types.CompletionItem(
                    maybe_quote_yaml_value(p)
                    if has_colon
                    else f"{maybe_quote_yaml_value(p)}:"
                )
                for p in binary_packages
                if p not in parent
            ]
    elif isinstance(parser, DeclarativeMappingInputParser):
        if matched_key:
            _info("Match attributes")
            locked = set(parent)
            for mx in parser.mutually_exclusive_attributes:
                if not mx.isdisjoint(parent.keys()):
                    locked.update(mx)
            for attr_name, attr in parser.manifest_attributes.items():
                if not attr.conflicting_attributes.isdisjoint(parent.keys()):
                    locked.add(attr_name)
                    break
            items = [
                types.CompletionItem(
                    maybe_quote_yaml_value(k)
                    if has_colon
                    else f"{maybe_quote_yaml_value(k)}:"
                )
                for k in parser.manifest_attributes
                if k not in locked
            ]
        else:
            # Value
            key = segments[at_depth_idx] if len(segments) > at_depth_idx else None
            value_attr = (
                parser.manifest_attributes.get(key) if isinstance(key, str) else None
            )
            if value_attr is not None:
                _info(f"Expand value / key: {key} -- {value_attr.attribute_type}")
                return completion_from_attr(
                    value_attr,
                    feature_set.manifest_parser_generator,
                    matched,
                    matched_key=False,
                    has_colon=has_colon,
                )
            else:
                _info(
                    f"Expand value / key: {key} -- !! {list(parser.manifest_attributes)}"
                )
    elif isinstance(parser, DeclarativeNonMappingInputParser):
        alt_attr = parser.alt_form_parser
        return completion_from_attr(
            alt_attr,
            feature_set.manifest_parser_generator,
            matched,
            matched_key=matched_key,
            has_colon=has_colon,
        )
    return items


def generic_yaml_hover(
    ls: "DebputyLanguageServer",
    params: types.HoverParams,
    root_parser_initializer: Callable[
        [ParserGenerator], DeclarativeInputParser[Any] | DispatchingParserBase
    ],
    *,
    show_integration_mode: bool = False,
) -> types.Hover | None:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lines = doc.lines
    position_codec = doc.position_codec
    server_position = position_codec.position_from_client_units(lines, params.position)

    try:
        content = MANIFEST_YAML.load(doc.source)
    except YAMLError:
        return None
    attribute_root_path = AttributePath.root_path(content)
    m = _trace_cursor(content, attribute_root_path, server_position)
    if m is None:
        _info("No match")
        return None
    matched_key, attr_path, matched, _ = m
    _info(f"Matched path: {matched} (path: {attr_path.path}) [{matched_key=}]")

    feature_set = ls.plugin_feature_set
    parser_generator = feature_set.manifest_parser_generator
    root_parser = root_parser_initializer(parser_generator)
    segments = list(attr_path.path_segments())
    km = resolve_keyword(
        root_parser,
        DEBPUTY_PLUGIN_METADATA,
        segments,
        0,
        parser_generator,
    )
    if km is None:
        _info("No keyword match")
        return None
    parser, plugin_metadata, at_depth_idx = km

    manifest_variable_at_pos = _manifest_substitution_variable_at_position(
        lines, server_position
    )

    if manifest_variable_at_pos:
        variable = ls.plugin_feature_set.manifest_variables.get(
            manifest_variable_at_pos
        )
        if variable is not None:
            var_doc = (
                variable.variable_reference_documentation
                or "No documentation available"
            )

            if variable.is_context_specific_variable:
                value = "\nThe value depends on the context"
            else:
                debian_dir = ls.lint_state(doc).debian_dir
                value = ""
                if debian_dir:
                    variable_context = VariableContext(debian_dir)
                    try:
                        resolved = variable.resolve(variable_context)
                        value = f"\nResolves to: `{resolved}`"
                    except RuntimeError:
                        pass

            hover_doc_text = textwrap.dedent("""\
            # `{NAME}`

            {DOC}
            {VALUE}
            """).format(NAME=variable.variable_name, DOC=var_doc, VALUE=value)
            return as_hover_doc(ls, hover_doc_text)

    _info(
        f"Match leaf parser {at_depth_idx}/{len(segments)} -- {parser.__class__} -- {manifest_variable_at_pos}"
    )
    hover_doc_text = resolve_hover_text(
        feature_set,
        parser,
        plugin_metadata,
        ls.hover_output_style,
        show_integration_mode,
        segments,
        at_depth_idx,
        matched,
        matched_key,
    )
    return as_hover_doc(ls, hover_doc_text)



import textwrap
from typing import TYPE_CHECKING

from debputy.lsp.languages.lsp_debian_changelog import _debian_changelog_links
from lint_tests.lint_tutil import lsp_range_to_text

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types

from lsp_tests.lsp_tutil import put_doc_no_cursor

try:
    from pygls.server import LanguageServer

    from debputy.lsp.languages.lsp_debian_rules import debian_rules_completions
    from debputy.lsp.debputy_ls import DebputyLanguageServer

    HAS_PYGLS = True
except ImportError:
    HAS_PYGLS = False


def _doc_link_req(uri: str) -> types.DocumentLinkParams:
    return types.DocumentLinkParams(types.TextDocumentIdentifier(uri))


def test_basic_debian_changelog_bug_links(ls: "DebputyLanguageServer") -> None:
    file_uri = "file:///nowhere/debian/changelog"
    text = textwrap.dedent("""\
        debputy (0.1.67) unstable; urgency=medium

          * debputy: some entry without bug links
          * debputy: Fix crash on creating dbgsym packages and other cases,
            where `debputy` would include a symlink in the package without
            a FS backed symlink behind it.
            Thanks to Andrea Pappacoda <tachi@debian.org> (Closes: #1100154)
          * This entry closes many bugs but this line does not have any.
            closes: bug 123456, # 98765
          * Also, you might be interested to see #564737 for this entry.

         -- Niels Thykier <niels@thykier.net>  Thu, 13 Mar 2025 16:50:42 +0000
        """)

    put_doc_no_cursor(
        ls,
        file_uri,
        "debian/changelog",
        text,
    )
    lines = text.splitlines(keepends=True)

    links = _debian_changelog_links(
        ls,
        _doc_link_req(file_uri),
    )
    print(links)
    assert links is not None
    assert len(links) == 4
    link_1, link_2, link_3, link_4 = links

    assert lsp_range_to_text(lines, link_1.range) == "#1100154"
    assert link_1.target == "https://bugs.debian.org/1100154"

    assert lsp_range_to_text(lines, link_2.range) == "bug 123456"
    assert link_2.target == "https://bugs.debian.org/123456"

    assert lsp_range_to_text(lines, link_3.range) == "# 98765"
    assert link_3.target == "https://bugs.debian.org/98765"

    assert lsp_range_to_text(lines, link_4.range) == "#564737"
    assert link_4.target == "https://bugs.debian.org/564737"



import textwrap
from functools import lru_cache
from typing import (
    Optional,
    Any,
    Union,
    TYPE_CHECKING,
    Type,
)
from collections.abc import Sequence

from debputy.linting.lint_util import LintState
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_standard_handler,
    lsp_hover,
    lsp_completer,
    LanguageDispatchRule,
    SecondaryLanguage,
)
from debputy.lsp.lsp_generic_yaml import (
    generic_yaml_hover,
    LSPYAMLHelper,
    generic_yaml_lint,
    generic_yaml_completer,
)
from debputy.manifest_parser.base_types import (
    DebputyParsedContent,
)
from debputy.manifest_parser.declarative_parser import (
    ParserGenerator,
)
from debputy.manifest_parser.parser_data import ParserContextData
from debputy.manifest_parser.util import AttributePath
from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
from debputy.plugin.api.impl_types import (
    DispatchingParserBase,
    DebputyPluginMetadata,
    DispatchingObjectParser,
)
from debputy.plugin.api.spec import ParserDocumentation, reference_documentation
from debputy.util import T

try:
    from debputy.lsp.debputy_ls import DebputyLanguageServer
    from debian._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )
except ImportError:
    pass

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types


_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/upstream/metadata",
    None,
    "debian/upstream/metadata",
    [SecondaryLanguage("yaml", secondary_lookup="path-name")],
)


lsp_standard_handler(_DISPATCH_RULE, types.TEXT_DOCUMENT_CODE_ACTION)
lsp_standard_handler(_DISPATCH_RULE, types.TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)


TT = type[T]


def _parser_handler(
    _key: str,
    value: Any,
    _attr_path: AttributePath,
    _context: Optional["ParserContextData"],
) -> Any:
    return value


def add_keyword(
    pg: ParserGenerator,
    root_parser: DispatchingParserBase[Any],
    plugin_metadata: DebputyPluginMetadata,
    keyword: str,
    value_type: TT,
    *,
    inline_reference_documentation: ParserDocumentation | None = None,
) -> None:
    class DebputyParsedContentWrapper(DebputyParsedContent):
        content: value_type  # type: ignore

    parser = pg.generate_parser(
        DebputyParsedContentWrapper,
        source_content=value_type,
        inline_reference_documentation=inline_reference_documentation,
    )
    root_parser.register_parser(
        keyword,
        parser,
        _parser_handler,
        plugin_metadata,
    )


@lru_cache
def root_object_parser() -> DispatchingObjectParser:
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    pg = ParserGenerator()
    root_parser = pg.add_object_parser(
        "<ROOT>",
        unknown_keys_diagnostic_severity="warning",
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Archive",
        str,
        inline_reference_documentation=reference_documentation(
            title="Archive (`Archive`)",
            description=textwrap.dedent("""\
                The name of the large archive that the upstream work is part of, like CPAN.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "ASCL-Id",
        str,
        inline_reference_documentation=reference_documentation(
            title="ASCL Identifier (`ASCL-Id`)",
            description=textwrap.dedent("""\
                Identification code in the http://ascl.net
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Bug-Database",
        str,
        inline_reference_documentation=reference_documentation(
            title="Bug database or tracker for the project (`Bug-Database`)",
            description=textwrap.dedent("""\
                A URL to the list of known bugs for the project.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Bug-Submit",
        str,
        inline_reference_documentation=reference_documentation(
            title="Bug submission URL for the project (`Bug-Submit`)",
            description=textwrap.dedent("""\
                A URL that is the place where new bug reports should be sent.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Cite-As",
        str,
        inline_reference_documentation=reference_documentation(
            title="Cite-As (`Cite-As`)",
            description=textwrap.dedent("""\
                The way the authors want their software be cited in publications.

                The value is a string which might contain a link in valid HTML syntax.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Changelog",
        str,
        inline_reference_documentation=reference_documentation(
            title="Changelog (`Changelog`)",
            description=textwrap.dedent("""\
                URL to the upstream changelog.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Contact",
        str,
        inline_reference_documentation=reference_documentation(
            title="Contact (`Contact`)",
            description=textwrap.dedent("""\
                Contact point for the upstream point.

                Deprecated when using the machine readable format. The `Upstream-Contact` field in
                `debian/copyright` is the direct replacement in that case.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "CPE",
        str,
        inline_reference_documentation=reference_documentation(
            title="CPE (`CPE`)",
            description=textwrap.dedent("""\
                One or more space separated http://cpe.mitre.org/ values useful to look up relevant CVEs
                in the https://nvd.nist.gov/home.cfm and other CVE sources.

                See `CPEtagPackagesDep` for information on how this information can be used.
                **Example**: `cpe:/a:ethereal_group:ethereal`
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Documentation",
        str,
        inline_reference_documentation=reference_documentation(
            title="Documentation (`Documentation`)",
            description=textwrap.dedent("""\
                A URL to online documentation.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Donation",
        str,
        inline_reference_documentation=reference_documentation(
            title="Donation (`Donation`)",
            description=textwrap.dedent("""\
                A URL to a donation form (or instructions).
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "FAQ",
        str,
        inline_reference_documentation=reference_documentation(
            title="FAQ (`FAQ`)",
            description=textwrap.dedent("""\
                A URL to the online FAQ.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Funding",
        # Unsure of the format
        Any,
        inline_reference_documentation=reference_documentation(
            title="Funding (`Funding`)",
            description=textwrap.dedent("""\
                One or more sources of funding which have supported this project (e.g. NSF OCI-12345).
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Gallery",
        str,
        inline_reference_documentation=reference_documentation(
            title="Gallery (`Gallery`)",
            description=textwrap.dedent("""\
                A URL to a gallery of pictures made with the program (not screenshots).
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Name",
        str,
        inline_reference_documentation=reference_documentation(
            title="Name (`Name`)",
            description=textwrap.dedent("""\
                Name of the upstream project.

                Deprecated when using the machine readable format. The `Upstream-Name` field in
                `debian/copyright` is the direct replacement in that case.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Other-References",
        str,
        inline_reference_documentation=reference_documentation(
            title="Other-References (`Other-References`)",
            description=textwrap.dedent("""\
                A URL to a upstream page containing more references.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Reference",
        # Complex type
        Any,
        inline_reference_documentation=reference_documentation(
            title="Reference (`Reference`)",
            # FIXME: Add the fields below as a nested subobject or list of such objects
            description=textwrap.dedent("""\
                One or more bibliographic references, represented as a mapping or sequence of mappings containing
                the one or more of the following keys.
 
                The values for the keys are always scalars, and the keys that correspond to standard BibTeX
                entries must provide the same content.
            """),
        ),
    )

    # Reference:: One or more bibliographic references, represented as a mapping or sequence of mappings containing the one or more of the following keys. The values for the keys are always scalars, and the keys that correspond to standard BibTeX entries must provide the same content.
    #
    # Author:: Author list in BibTeX friendly syntax (separating multiple authors by the keyword "and" and using as few as possible abbreviations in the names, as proposed in http://nwalsh.com/tex/texhelp/bibtx-23.html).
    #
    # Booktitle:: Title of the book the article is published in
    #
    # DOI:: This is the digital object identifier of the academic publication describing the packaged work.
    #
    # Editor:: Editor of the book the article is published in
    #
    # Eprint:: Hyperlink to the PDF file of the article.
    #
    # ISBN:: International Standard Book Number of the book if the article is part of the book or the reference is a book
    #
    # ISSN:: International Standard Serial Number of the periodical publication if the article is part of a series
    #
    # Journal:: Abbreviated journal name [To be discussed: which standard to recommend ?].
    #
    # Number:: Issue number.
    #
    # Pages:: Article page number(s). [To be discussed] Page number separator must be a single ASCII hyphen. What do we do with condensed notations like 401-10 ?
    #
    # PMID:: ID number in the https://www.ncbi.nlm.nih.gov/pubmed/ database.
    #
    # Publisher:: Publisher of the book containing the article
    #
    # Title:: Article title.
    #
    # Type:: A http://www.bibtex.org/Format indicating what is cited. Typical values are {{{article}}}, {{{book}}}, or {{{inproceedings}}}. [To be discussed]. In case this field is not present, {{{article}}} is assumed.
    #
    # URL:: Hyperlink to the abstract of the article. This should not point to the full version because this is specified by Eprint. Please also do not drop links to pubmed here because this would be redundant to PMID.
    #
    # Volume:: Journal volume.
    #
    # Year:: Year of publication
    #

    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Registration",
        str,
        inline_reference_documentation=reference_documentation(
            title="Registration (`Registration`)",
            description=textwrap.dedent("""\
                A URL to a registration form (or instructions). This could be registration of bug reporting
                accounts, registration for counting/contacting users etc.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Registry",
        # FIXME: Add List of `Name`, `Entry` objects
        Any,
        inline_reference_documentation=reference_documentation(
            title="Registry (`Registry`)",
            description=textwrap.dedent("""\
                    This field shall point to external catalogs/registries of software.
 
                    The field features an array of "Name (of registry) - Entry (ID of software in that catalog)" pairs.
                    The names and entries shall only be names, not complete URIs, to avoid any bias on mirrors etc.
                    Example:
                    ```yaml
                       Registry:
                         - Name: bio.tools
                           Entry: clustalw
                         - Name: OMICtools
                           Entry: OMICS_02562
                         - Name: SciCrunch
                           Entry: SCR_002909
                    ```
            """),
        ),
    )

    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Repository",
        str,
        inline_reference_documentation=reference_documentation(
            title="Repository (`Repository`)",
            description=textwrap.dedent("""\
                    URL to a repository containing the upstream sources.
            """),
        ),
    )

    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Repository-Browse",
        str,
        inline_reference_documentation=reference_documentation(
            title="Repository-Browse (`Repository-Browse`)",
            description=textwrap.dedent("""\
                     A URL to browse the repository containing the upstream sources.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Screenshots",
        Any,
        inline_reference_documentation=reference_documentation(
            title="Screenshots (`Screenshots`)",
            description=textwrap.dedent("""\
                     One or more URLs to upstream pages containing screenshots (not <https://screenshots.debian.net>),
                     represented by a scalar or a sequence of scalars.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Security-Contact",
        str,
        inline_reference_documentation=reference_documentation(
            title="Security-Contact (`Security-Contact`)",
            description=textwrap.dedent("""\
                      Which person, mailing list, forum, etc. to send security-related messages in the first place.
            """),
        ),
    )
    add_keyword(
        pg,
        root_parser,
        plugin_metadata,
        "Webservice",
        str,
        inline_reference_documentation=reference_documentation(
            title="Webservice (`Webservice`)",
            description=textwrap.dedent("""\
                      URL to a web page where the packaged program can also be used.
            """),
        ),
    )
    return root_parser


def _initialize_yaml_helper(lint_state: LintState) -> LSPYAMLHelper[None]:
    return LSPYAMLHelper(
        lint_state,
        lint_state.plugin_feature_set.manifest_parser_generator,
        None,
    )


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_upstream_metadata(lint_state: LintState) -> None:
    await generic_yaml_lint(
        lint_state,
        root_object_parser(),
        _initialize_yaml_helper,
    )


@lsp_completer(_DISPATCH_RULE)
def debian_upstream_metadata_completer(
    ls: "DebputyLanguageServer",
    params: types.CompletionParams,
) -> types.CompletionList | Sequence[types.CompletionItem] | None:
    return generic_yaml_completer(
        ls,
        params,
        root_object_parser(),
    )


@lsp_hover(_DISPATCH_RULE)
def debputy_manifest_hover(
    ls: "DebputyLanguageServer",
    params: types.HoverParams,
) -> types.Hover | None:
    return generic_yaml_hover(ls, params, lambda _: root_object_parser())



from fileinput import filename
from typing import TextIO

class TestCase:
    def __init__(
        self,
        filename: str,
        *,
        # The JUnit schema has `classname` as mandatory
        classname: str,
        allow_multiple_subelements: bool,
        elapsed_sec: float,
    ) -> None: ...
    def add_failure_info(
        self,
        *,
        message: str,
        output: str,
    ) -> None: ...

class TestSuite:
    def __init__(
        self,
        name: str,
        *,
        test_cases: list[TestCase],
        timestamp: str,
    ) -> None: ...

def to_xml_report_file(
    fd: TextIO,
    test_suites: list[TestSuite],
    *,
    encoding: str,
) -> None: ...



import textwrap

import pytest

from debputy.lsprotocol.types import DiagnosticSeverity

from debputy.lsp.languages.lsp_debian_watch import _lint_debian_watch
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from lint_tests.lint_tutil import (
    LintWrapper,
)


@pytest.fixture
def line_linter(
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
    lint_dctrl_parser: DctrlParser,
) -> LintWrapper:
    return LintWrapper(
        "/nowhere/debian/watch",
        _lint_debian_watch,
        debputy_plugin_feature_set,
        lint_dctrl_parser,
    )


def test_dwatch_alias_dup(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent("""\
    Version: 5

    Source: https://example.org
    MatchingPattern: @ANY_VERSION@
    Versionregex: @ANY_VERSION@
    """).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    diagnostics = [d for d in diagnostics if d.severity == DiagnosticSeverity.Error]
    assert len(diagnostics) == 1
    diag = diagnostics[0]

    msg = 'The field "Matching-Pattern" (via aliases) was used multiple times in this stanza. Please ensure the field is only used once per stanza.'
    assert diag.message == msg
    assert f"{diag.range}" == "4:0-4:12"



import textwrap

import pytest

from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.lsp.languages.lsp_debian_watch import _debian_watch_hover
from debputy.lsprotocol.types import (
    TextDocumentIdentifier,
    HoverParams,
    MarkupContent,
)

try:
    from debputy.lsp.languages.lsp_debian_control import (
        _debian_control_completions,
        _debian_control_hover,
        _debian_control_semantic_tokens_full,
    )

    from pygls.server import LanguageServer
except ImportError:
    pass
from lsp_tests.lsp_tutil import (
    put_doc_with_cursor,
)


@pytest.mark.parametrize(
    "variant",
    [
        "foo-<CURSOR>@ANY_VERSION@",
        "foo-@<CURSOR>ANY_VERSION@",
        "foo-@ANY_VERSION<CURSOR>@",
    ],
)
def test_debian_watch_hover_doc_variables_matching(
    ls: "DebputyLanguageServer", variant: str
) -> None:
    dwatch_uri = "file:///nowhere/debian/watch"
    cursor_pos = put_doc_with_cursor(
        ls,
        dwatch_uri,
        "debian/watch",
        textwrap.dedent(f"""\
        Version: 5

        Source: https://example.org
        Matching-Pattern: {variant}
"""),
    )

    hover_doc = _debian_watch_hover(
        ls,
        HoverParams(TextDocumentIdentifier(dwatch_uri), cursor_pos),
    )
    assert hover_doc is not None and isinstance(hover_doc.contents, MarkupContent)
    assert hover_doc.contents.value.startswith("# Variable `@ANY_VERSION@`")


@pytest.mark.parametrize(
    "variant",
    [
        "foo<CURSOR>-@ANY_VERSION@",
        "foo-@ANY_VERSION@<CURSOR>",
    ],
)
def test_debian_watch_hover_doc_variables_non_matching(
    ls: "DebputyLanguageServer", variant: str
) -> None:
    dwatch_uri = "file:///nowhere/debian/watch"
    cursor_pos = put_doc_with_cursor(
        ls,
        dwatch_uri,
        "debian/watch",
        textwrap.dedent(f"""\
                Version: 5

                Source: https://example.org
                Matching-Pattern: {variant}
        """),
    )

    hover_doc = _debian_watch_hover(
        ls,
        HoverParams(TextDocumentIdentifier(dwatch_uri), cursor_pos),
    )
    provided_doc = ""
    if hover_doc is not None and isinstance(hover_doc.contents, MarkupContent):
        provided_doc = hover_doc.contents.value
    assert not provided_doc.startswith("# Variable `@ANY_VERSION@`")



from .colored import Colored as Colored

def cprint(
    text: str,
    fore_256: int | str = "",
    back_256: int | str = "",
    fore_rgb: tuple = (255, 255, 255),
    back_rgb: tuple = (0, 0, 0),
    formatting: int | str = "",
    line_color: int | str = "",
    reset: bool = True,
    **kwargs,
) -> None: ...



import textwrap
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types

from lsp_tests.lsp_tutil import put_doc_with_cursor

try:
    from pygls.server import LanguageServer

    from debputy.lsp.languages.lsp_debian_rules import debian_rules_completions
    from debputy.lsp.debputy_ls import DebputyLanguageServer

    HAS_PYGLS = True
except ImportError:
    HAS_PYGLS = False


def _complete_at_req(uri: str, cursor_pos: types.Position) -> types.CompletionParams:
    return types.CompletionParams(types.TextDocumentIdentifier(uri), cursor_pos)


def test_basic_debian_rules_completer_include(ls: "DebputyLanguageServer") -> None:
    file_uri = "file:///nowhere/debian/rules"

    cursor_pos = put_doc_with_cursor(
        ls,
        file_uri,
        "debian/rules",
        "include <CURSOR>",
    )

    completions = debian_rules_completions(
        ls,
        _complete_at_req(file_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label: m for m in completions}
    assert "/usr/share/dpkg/default.mk" in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        file_uri,
        "debian/rules",
        textwrap.dedent("""\
        include /usr/share/<CURSOR>dpkg/default.m
"""),
    )

    completions = debian_rules_completions(
        ls,
        _complete_at_req(file_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label: m for m in completions}
    assert "/usr/share/dpkg/default.mk" in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        file_uri,
        "debian/rules",
        textwrap.dedent("""\
        include /usr/share/dpkg/default.mk<CURSOR>
"""),
    )

    completions = debian_rules_completions(
        ls,
        _complete_at_req(file_uri, cursor_pos),
    )
    # No completions (it matches a known value)
    assert not completions

    cursor_pos = put_doc_with_cursor(
        ls,
        file_uri,
        "debian/rules",
        textwrap.dedent("""\
        include /usr foo<CURSOR>
"""),
    )

    completions = debian_rules_completions(
        ls,
        _complete_at_req(file_uri, cursor_pos),
    )
    # No completions (it has a space, so something funky is going on)
    assert not completions



import dataclasses
import textwrap
from typing import (
    Optional,
    Union,
    Any,
    Self,
    TYPE_CHECKING,
)
from collections.abc import Mapping, Sequence, Callable, Iterable

from debputy.lsp.named_styles import ALL_PUBLIC_NAMED_STYLES
from debputy.lsp.ref_models.deb822_reference_parse_models import UsageHint
from debian._deb822_repro import Deb822ParagraphElement

if TYPE_CHECKING:
    import lsprotocol.types as types
    from debputy.linting.lint_util import LintState
    from debputy.lsp.debputy_ls import DebputyLanguageServer
else:
    import debputy.lsprotocol.types as types

LSP_DATA_DOMAIN = "debputy-lsp-data"


def format_comp_item_synopsis_doc(
    usage_hint: UsageHint | None,
    synopsis_doc: str | None,
    is_deprecated: bool,
) -> str:
    if is_deprecated:
        return (
            f"[OBSOLETE]: {synopsis_doc}"
            if synopsis_doc is not None and not synopsis_doc.isspace()
            else f"[OBSOLETE]"
        )
    if usage_hint is not None:
        return (
            f"[{usage_hint.upper()}]: {synopsis_doc}"
            if synopsis_doc is not None and not synopsis_doc.isspace()
            else f"[{usage_hint.upper()}]"
        )
    return synopsis_doc or ""


@dataclasses.dataclass(slots=True, frozen=True)
class Keyword:
    value: str
    synopsis: str | None = None
    long_description: str | None = None
    translation_context: str = ""
    is_obsolete: bool = False
    replaced_by: str | None = None
    is_exclusive: bool = False
    """For keywords in fields that allow multiple keywords, the `is_exclusive` can be
    used for keywords that cannot be used with other keywords. As an example, the `all`
    value in `Architecture` of `debian/control` cannot be used with any other architecture.
    """
    is_alias_of: str | None = None
    is_completion_suggestion: bool = True
    sort_text: None | (
        str
        | Callable[["Keyword", "LintState", Sequence[Deb822ParagraphElement], str], str]
    ) = None
    usage_hint: UsageHint | None = None
    can_complete_keyword_in_stanza: None | (
        Callable[[Iterable[Deb822ParagraphElement]], bool]
    ) = None

    @property
    def is_deprecated(self) -> bool:
        return self.is_obsolete or self.replaced_by is not None

    def resolve_sort_text(
        self,
        lint_state: "LintState",
        stanza_parts: Sequence[Deb822ParagraphElement],
        value_being_completed: str,
    ) -> str | None:
        sort_text = self.sort_text
        if sort_text is None:
            return None
        if not isinstance(sort_text, str):
            return sort_text(
                self,
                lint_state,
                stanza_parts,
                value_being_completed,
            )
        return sort_text

    def is_keyword_valid_completion_in_stanza(
        self,
        stanza_parts: Sequence[Deb822ParagraphElement],
    ) -> bool:
        return (
            self.can_complete_keyword_in_stanza is None
            or self.can_complete_keyword_in_stanza(stanza_parts)
        )

    def replace(self, **changes: Any) -> "Self":
        return dataclasses.replace(self, **changes)

    def synopsis_translated(
        self, translation_provider: Union["DebputyLanguageServer", "LintState"]
    ) -> str | None:
        if self.synopsis is None:
            return None
        return translation_provider.translation(LSP_DATA_DOMAIN).pgettext(
            self.translation_context,
            self.synopsis,
        )

    def long_description_translated(
        self, translation_provider: Union["DebputyLanguageServer", "LintState"]
    ) -> str | None:
        if self.long_description is None:
            return None
        return translation_provider.translation(LSP_DATA_DOMAIN).pgettext(
            self.translation_context,
            self.long_description,
        )

    def as_completion_item(
        self,
        lint_state: "LintState",
        stanza_parts: Sequence[Deb822ParagraphElement],
        value_being_completed: str,
        markdown_kind: types.MarkupKind,
    ) -> types.CompletionItem:
        return types.CompletionItem(
            self.value,
            insert_text=self.value if self.is_alias_of is None else self.is_alias_of,
            sort_text=self.resolve_sort_text(
                lint_state,
                stanza_parts,
                value_being_completed,
            ),
            detail=format_comp_item_synopsis_doc(
                self.usage_hint,
                self.synopsis_translated(lint_state),
                self.is_deprecated,
            ),
            deprecated=self.is_deprecated,
            tags=[types.CompletionItemTag.Deprecated] if self.is_deprecated else None,
            documentation=(
                types.MarkupContent(value=self.long_description, kind=markdown_kind)
                if self.long_description
                else None
            ),
        )


def allowed_values(values: Iterable[str | Keyword]) -> Mapping[str, Keyword]:
    as_keywords = [k if isinstance(k, Keyword) else Keyword(k) for k in values]
    as_mapping = {k.value: k for k in as_keywords if k.value}
    # Simple bug check
    assert len(as_keywords) == len(as_mapping)
    return as_mapping


# This is the set of styles that `debputy` explicitly supports, which is more narrow than
# the ones in the config file.
ALL_PUBLIC_NAMED_STYLES_AS_KEYWORDS = allowed_values(
    Keyword(s.name, long_description=s.long_description)
    for s in ALL_PUBLIC_NAMED_STYLES.values()
)



from .attributes import Style as Style
from .background import Back as Back
from .colored import (
    Colored as Colored,
    attr as attr,
    back as back,
    back_rgb as back_rgb,
    bg as bg,
    fg as fg,
    fore as fore,
    fore_rgb as fore_rgb,
    set_tty_aware as set_tty_aware,
    style as style,
    stylize as stylize,
    stylize_interactive as stylize_interactive,
)
from .controls import Controls as Controls
from .cprint import cprint as cprint
from .foreground import Fore as Fore

__version__: str



import textwrap
from typing import Optional, Type, Any

from debputy.manifest_parser.declarative_parser import (
    ParserGenerator,
    DeclarativeNonMappingInputParser,
)
from debputy.manifest_parser.tagging_types import DebputyParsedContent
from debputy.plugin.api import reference_documentation
from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
from debputy.plugin.api.impl_types import (
    DispatchingObjectParser,
    DebputyPluginMetadata,
    DeclarativeInputParser,
)
from debputy.plugin.api.spec import ParserDocumentation
from debputy.util import T, _error
from .config_options import ALL_DEBPUTY_CONFIG_OPTIONS


class BoolDebputyParsedContent(DebputyParsedContent):
    content: bool


TT = type[T]


def _debputy_config_parser() -> DispatchingObjectParser:
    pg = ParserGenerator()
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    root_parser = pg.add_object_parser(
        "<ROOT>",
        unknown_keys_diagnostic_severity=None,
        allow_unknown_keys=True,
    )
    diagnostics_object_parser = _add_sub_object_parser(
        pg,
        root_parser,
        "diagnostics",
        plugin_metadata,
        parser_documentation=reference_documentation(
            title="Diagnostics related settings (`diagnostics`)",
            description=textwrap.dedent("""\
                Configurations related to diagnostics.
            """),
        ),
    )

    spellchecking_diagnostics = _add_sub_object_parser(
        pg,
        diagnostics_object_parser,
        "spellchecking",
        plugin_metadata,
        parser_documentation=reference_documentation(
            title="Spellchecking related diagnostics settings (`spellchecking`)",
            description=textwrap.dedent("""\
                Settings related to spellchecking diagnostics
            """),
        ),
    )
    _config_value(
        pg,
        spellchecking_diagnostics,
        "spellcheck-comments",
        bool,
        plugin_metadata,
        inline_reference_documentation=reference_documentation(
            title="Whether to spellcheck comments (`spellcheck-comments`)",
            description=textwrap.dedent("""\
                Define whether `debputy` should spellcheck syntactical comments.

                Consider the following example:
                ```debian/control
                Source: foo

                Package: foo
                # Here is a comment with a typu.
                Description: Also a mistkae here
                ```

                This option affects whether `typu` in the above example will be
                reported provided textual spellchecking is enabled. The other typo
                (`mistkae`) is unaffected by this option, since it is not inside
                a syntactical comment.

                When this option is present and set to `false`, then syntactical comments
                are never spellchecked.
            """),
        ),
    )
    _debputy_self_check_config_parser(root_parser)
    return root_parser


def _debputy_self_check_config_parser(root_parser: DispatchingObjectParser) -> None:
    for config_option in ALL_DEBPUTY_CONFIG_OPTIONS:
        parser: DeclarativeInputParser[Any] = root_parser
        for part in config_option.config_name.split("."):
            if isinstance(parser, DispatchingObjectParser):
                parser = parser.parser_for(part).parser
            else:
                _error(
                    f"Unknown parser for {config_option.config_name} at {part}: {parser}"
                )
        if not isinstance(parser, DeclarativeNonMappingInputParser):
            _error(
                f"Expecting {config_option.config_name} to resolve to a DeclarativeNonMappingInputParser, got: {parser}"
            )
        if parser.alt_form_parser.attribute_type != config_option.value_type:
            _error(
                f"Expecting {config_option.config_name} to be of type {config_option.value_type}, but parser suggest it is {parser.alt_form_parser.attribute_type}"
            )


def _config_value(
    pg: ParserGenerator,
    parent_parser: DispatchingObjectParser,
    path: str,
    value_type: TT,
    plugin_metadata: DebputyPluginMetadata,
    *,
    inline_reference_documentation: ParserDocumentation | None = None,
) -> None:
    class DebputyParsedContentWrapper(DebputyParsedContent):
        content: value_type  # type: ignore

    parser = pg.generate_parser(
        DebputyParsedContentWrapper,
        source_content=value_type,
        inline_reference_documentation=inline_reference_documentation,
    )
    parent_parser.register_parser(
        path,
        parser,
        lambda k, pf, ap, pc: pf["content"],
        plugin_metadata,
    )


def _add_sub_object_parser(
    pg: ParserGenerator,
    parent_parser: DispatchingObjectParser,
    path: str,
    plugin_metadata: DebputyPluginMetadata,
    *,
    parser_documentation: ParserDocumentation | None = None,
) -> DispatchingObjectParser:
    if parent_parser.manifest_attribute_path_template == "<ROOT>":
        full_path = path
    else:
        full_path = f"{parent_parser.manifest_attribute_path_template}.{path}"
    child_parser = pg.add_object_parser(
        full_path,
        unknown_keys_diagnostic_severity=None,
        parser_documentation=parser_documentation,
        allow_unknown_keys=True,
    )
    parent_parser.register_parser(
        path,
        child_parser,
        lambda k, pf, ap, pc: pf,
        plugin_metadata,
    )
    return child_parser


DEBPUTY_CONFIG_PARSER = _debputy_config_parser()

del _debputy_config_parser
del _config_value
del _add_sub_object_parser
del _debputy_self_check_config_parser



"debputy: test generation of the (static-)built-using field."

import typing

import pytest

from conftest import ManifestParserFactory
from debputy.filesystem_scan import InMemoryVirtualRootDir
from debputy.plugin.api import virtual_path_def
from debputy.plugin.api.impl_types import (
    PackageProcessingContextProvider,
    PackageDataTable,
)
from debputy.plugin.api.test_api import build_virtual_file_system
from debputy.plugin.api.test_api.test_impl import initialize_plugin_under_test_preloaded
from debputy.plugins.debputy.debputy_plugin import initialize_debputy_features
from debputy.util import PKGNAME_REGEX
from tutil import is_pkg_installed


class TDefinition(typing.NamedTuple):
    name: str
    manifest: str
    expected: set[str]
    build_depends: str
    architecture: str = "any"
    field: str = "Built-Using"
    required_pkgs: tuple[str, ...] = ()


# More or less in sync with dh-builtusing unit-tests.
tests = (
    TDefinition(
        name="01basic",
        build_depends="dpkg",
        manifest="sources-for: dpkg",
        expected={"dpkg"},
    ),
    TDefinition(
        name="01static",
        build_depends="dpkg",
        manifest="sources-for: dpkg",
        expected={"dpkg"},
        field="Static-Built-Using",
    ),
    TDefinition(
        name="20build-depends-arch-restriction",
        build_depends="dpkg [i386]",
        manifest="sources-for: dpkg",
        expected={},
    ),
    TDefinition(
        name="20build-depends-profile-restriction",
        build_depends="dpkg <dummy>",
        manifest="sources-for: dpkg",
        expected={},
    ),
    TDefinition(
        name="20build-depends-arch",
        required_pkgs=("libc6", "libdpkg-perl", "libbinutils"),
        build_depends="""libc6
Build-Depends-Arch: libdpkg-perl
Build-Depends-Indep: libbinutils""",
        manifest="sources-for: 'lib*'",
        expected={"glibc", "dpkg"},
    ),
    TDefinition(
        name="20build-depends-indep",
        required_pkgs=("libc6", "libdpkg-perl", "libbinutils"),
        architecture="all",
        build_depends="""libc6
Build-Depends-Arch: libdpkg-perl
Build-Depends-Indep: libbinutils""",
        manifest="sources-for: 'lib*'",
        expected={"glibc", "binutils"},
    ),
    TDefinition(
        name="30or-dependency-first",
        build_depends="dpkg | dummy",
        manifest="sources-for: dpkg",
        expected={"dpkg"},
    ),
    TDefinition(
        name="30or-dependency-not-first",
        build_depends="dummy | dpkg",
        manifest="sources-for: dpkg",
        expected={"dpkg"},
    ),
    TDefinition(
        name="40pattern-star-initial",
        build_depends="dpkg",
        manifest="sources-for: '*kg'",
        expected={"dpkg"},
    ),
    TDefinition(
        name="40pattern-star-final",
        build_depends="dpkg",
        manifest="sources-for: 'dp*'",
        expected={"dpkg"},
    ),
    TDefinition(
        name="40pattern-star-empty",
        build_depends="dpkg",
        manifest="sources-for: 'dp*kg'",
        expected={"dpkg"},
    ),
    TDefinition(
        name="40pattern-star-one-char",
        build_depends="dpkg",
        manifest="sources-for: 'dpk*g'",
        expected={"dpkg"},
    ),
    TDefinition(
        name="40pattern-re-char-plus",
        required_pkgs=("g++",),
        build_depends="g++",
        manifest="sources-for: g*+",
        expected={"gcc-defaults"},
    ),
    TDefinition(
        name="40pattern-ambiguous-all",
        required_pkgs=("libbinutils", "libc6"),
        build_depends="""
Build-Depends-Arch: libbinutils
Build-Depends-Indep: libc6""",
        architecture="all",
        manifest="sources-for: 'lib*'",
        expected={"glibc"},
    ),
    TDefinition(
        name="40pattern-ambiguous-any",
        required_pkgs=("libbinutils", "libc6"),
        build_depends="""
Build-Depends-Arch: libbinutils
Build-Depends-Indep: libc6""",
        manifest="sources-for: 'lib*'",
        expected={"binutils"},
    ),
    TDefinition(
        name="45multiple-matches",
        required_pkgs=("dpkg-dev", "autotools-dev"),
        build_depends="dpkg-dev, autotools-dev",
        manifest="sources-for: '*-dev'",
        expected={"autotools-dev", "dpkg"},
    ),
    TDefinition(
        name="50same-source",
        required_pkgs=("dpkg-dev",),
        build_depends="dpkg, dpkg-dev",
        manifest="sources-for: dpkg, sources-for: dpkg-dev",
        expected={"dpkg"},
    ),
    TDefinition(
        name="70arch-profile-restrictions-arch-yes",
        build_depends="dpkg",
        manifest="{sources-for: dpkg, when: {arch-matches: amd64}}",
        expected={"dpkg"},
    ),
    TDefinition(
        name="70arch-profile-restrictions-arch-no",
        build_depends="dpkg",
        manifest="{sources-for: dpkg, when: {arch-matches: '!amd64'}}",
        expected={},
    ),
    TDefinition(
        name="70arch-profile-restrictions-profile-no",
        build_depends="dpkg",
        manifest="{sources-for: dpkg, when: {build-profiles-matches: '<dummy>'}}",
        expected={},
    ),
    TDefinition(
        name="70arch-suffix-all",
        required_pkgs=("debhelper",),
        build_depends="debhelper:all",
        manifest="sources-for: debhelper",
        expected={"debhelper"},
    ),
    TDefinition(
        name="70arch-suffix-native",
        required_pkgs=("gcc",),
        build_depends="gcc:amd64",
        manifest="sources-for: gcc",
        expected={"gcc-defaults"},
    ),
    TDefinition(
        name="70arch-suffix-same",
        required_pkgs=("libc6",),
        build_depends="libc6:amd64",
        manifest="sources-for: libc6",
        expected={"glibc"},
    ),
    TDefinition(
        name="70multiarch-no",
        required_pkgs=("gcc",),
        build_depends="gcc",
        manifest="sources-for: gcc",
        expected={"gcc-defaults"},
    ),
    # 70multiarch-foreign would duplicate 01basic
    TDefinition(
        name="70multiarch-same",
        required_pkgs=("libc6",),
        build_depends="libc6",
        manifest="sources-for: libc6",
        expected={"glibc"},
    ),
    TDefinition(
        name="70multiarch-allowed",
        required_pkgs=("make",),
        build_depends="make",
        manifest="sources-for: make",
        expected={"make-dfsg"},
    ),
)


def _pkg_names_from_substvar(resolve_substvar: str) -> set[str]:
    return set(
        PKGNAME_REGEX.match(p.strip()).group(0) for p in resolve_substvar.split(",")
    )


@pytest.mark.parametrize("test_definition", tests, ids=[t.name for t in tests])
def test_built_using(
    manifest_parser_amd64_factory: ManifestParserFactory,
    test_definition: TDefinition,
) -> None:

    for p in test_definition.required_pkgs:
        if not is_pkg_installed(p):
            pytest.skip(f"The test requires the {p} package installed")
            return

    dctrl_contents = f"""\
Source: foo
Build-Depends: {test_definition.build_depends}

Package: foo
Architecture: {test_definition.architecture}
"""

    manifest_contents = f"""\
manifest-version: '0.1'
packages:
  foo:
    {test_definition.field.lower()}: [{test_definition.manifest}]
"""

    # A bit of debugging output in case the test fails
    print(" --- d/control ---")
    print(dctrl_contents)
    print(" --- d/debputy.manifest ---")
    print(manifest_contents)

    manifest = manifest_parser_amd64_factory(
        build_virtual_file_system(
            (virtual_path_def("./control", content=dctrl_contents),),
        )
    ).parse_manifest(fd=manifest_contents)
    binary_package = manifest.package_state_for("foo").binary_package
    assert binary_package.should_be_acted_on
    substvars = (
        initialize_plugin_under_test_preloaded(
            1,
            initialize_debputy_features,
            plugin_name="debputy",
            load_debputy_plugin=False,
        )
        .run_metadata_detector(
            metadata_detector_id="built-using-relations",
            fs_root=InMemoryVirtualRootDir(),
            context=PackageProcessingContextProvider(
                manifest=manifest,
                binary_package=binary_package,
                related_udeb_package=None,
                package_data_table=PackageDataTable({}),
            ),
        )
        .substvars
    )
    key = f"debputy:{test_definition.field}"
    if test_definition.expected:
        actual = _pkg_names_from_substvar(substvars[key])
        assert actual == test_definition.expected
    else:
        assert key not in substvars



import os
from typing import Any, Optional, cast, Dict
from collections.abc import Mapping

from debputy.lsp.config.config_options import (
    DebputyConfigOption,
    ALL_DEBPUTY_CONFIG_OPTIONS,
)
from debputy.lsp.config.parser import DEBPUTY_CONFIG_PARSER
from debputy.manifest_parser.util import AttributePath
from debputy.util import T
from debputy.yaml import MANIFEST_YAML


class DebputyConfig:

    def __init__(self, config: dict[DebputyConfigOption[Any], Any]) -> None:
        self._config = config

    def config_value(self, config_option: DebputyConfigOption[T]) -> T | None:
        return cast("Optional[T]", self._config.get(config_option))


def _resolve(parsed_config: Mapping[str, Any], config_name: str) -> Any | None:
    parts = config_name.split(".")
    container = parsed_config
    value: Any | None = None
    for part in parts:
        if isinstance(container, Mapping):
            value = container.get(part)
            container = value
        else:
            return None
    return value


def _read_option(
    parsed_config: Mapping[str, Any],
    config_option: DebputyConfigOption[T],
) -> T | None:
    result = _resolve(parsed_config, config_option.config_name)
    if result is None:
        return config_option.default_value
    assert isinstance(result, config_option.value_type)
    return result


def load_debputy_config() -> DebputyConfig:
    config_base = os.environ.get("XDG_CONFIG_DIR", os.path.expanduser("~/.config"))
    config_file = os.path.join(config_base, "debputy", "debputy-config.yaml")
    try:
        with open(config_file) as fd:
            parsed_yaml = MANIFEST_YAML.load(fd)
    except FileNotFoundError:
        parsed_yaml = None
    if not parsed_yaml:
        return DebputyConfig({})
    parsed_config = DEBPUTY_CONFIG_PARSER.parse_input(
        parsed_yaml,
        AttributePath.root_path(config_file),
        parser_context=None,
    )
    debputy_config = {}
    for config_option in ALL_DEBPUTY_CONFIG_OPTIONS:
        value = _read_option(parsed_config, config_option)
        debputy_config[config_option] = value
    return DebputyConfig(debputy_config)



"Access to DEB_BUILD_OPTIONS and DEB_BUILD_OPTIONS from the environment."

import functools
import os
from collections.abc import Mapping


@functools.cache
def _parse_profiles(value: str) -> frozenset[str]:
    return frozenset(value.split())


@functools.cache
def _parse_deb_build_options(value: str) -> Mapping[str, str | None]:
    res: dict[str, str | None] = {}
    for kvish in value.split():
        if "=" in kvish:
            key, value = kvish.split("=", 1)
            res[key] = value
        else:
            res[kvish] = None
    return res


class DebBuildOptionsAndProfiles:
    """Accessor to common environment related values

    >>> env = DebBuildOptionsAndProfiles(environ={'DEB_BUILD_PROFILES': 'noudeb nojava'})
    >>> 'noudeb' in env.deb_build_profiles
    True
    >>> 'nojava' in env.deb_build_profiles
    True
    >>> 'nopython' in env.deb_build_profiles
    False
    >>> sorted(env.deb_build_profiles)
    ['nojava', 'noudeb']
    """

    def __init__(self, *, environ: Mapping[str, str] | None = None) -> None:
        """Provide a view of the options.

        :param environ: Alternative to os.environ. Mostly useful for testing purposes
        """
        self._env = os.environ if environ is None else environ

    @property
    def deb_build_profiles(self) -> frozenset[str]:
        """A set-like view of all build profiles active during the build

        >>> env = DebBuildOptionsAndProfiles(environ={'DEB_BUILD_PROFILES': 'noudeb nojava'})
        >>> 'noudeb' in env.deb_build_profiles
        True
        >>> 'nojava' in env.deb_build_profiles
        True
        >>> 'nopython' in env.deb_build_profiles
        False
        >>> sorted(env.deb_build_profiles)
        ['nojava', 'noudeb']

        """
        return _parse_profiles(self._env.get("DEB_BUILD_PROFILES", ""))

    @property
    def deb_build_options(self) -> Mapping[str, str | None]:
        """A set-like view of all build profiles active during the build

        >>> env = DebBuildOptionsAndProfiles(environ={'DEB_BUILD_OPTIONS': 'nostrip parallel=4'})
        >>> 'nostrip' in env.deb_build_options
        True
        >>> 'parallel' in env.deb_build_options
        True
        >>> 'noautodbgsym' in env.deb_build_options
        False
        >>> env.deb_build_options['nostrip'] is None
        True
        >>> env.deb_build_options['parallel']
        '4'
        >>> env.deb_build_options['noautodbgsym']
        Traceback (most recent call last):
            ...
        KeyError: 'noautodbgsym'
        >>> sorted(env.deb_build_options)
        ['nostrip', 'parallel']

        """
        return _parse_deb_build_options(self._env.get("DEB_BUILD_OPTIONS", ""))



import itertools
from string import Template
from typing import Any, Union
from collections.abc import Iterable, Mapping, Sequence, Container

from debputy.commands.debputy_cmd.output import OutputStyle
from debputy.manifest_parser.declarative_parser import (
    DeclarativeMappingInputParser,
    DeclarativeNonMappingInputParser,
    AttributeDescription,
    BASIC_SIMPLE_TYPES,
)
from debputy.manifest_parser.parser_data import ParserContextData
from debputy.manifest_parser.tagging_types import TypeMapping
from debputy.manifest_parser.util import AttributePath, unpack_type
from debputy.plugin.api.impl_types import (
    DebputyPluginMetadata,
    DeclarativeInputParser,
    DispatchingObjectParser,
    ListWrappedDeclarativeInputParser,
    InPackageContextParser,
    PluginProvidedTypeMapping,
)
from debputy.plugin.api.spec import (
    ParserDocumentation,
    reference_documentation,
    undocumented_attr,
    DebputyIntegrationMode,
    ALL_DEBPUTY_INTEGRATION_MODES,
    TypeMappingExample,
)
from debputy.util import assume_not_none, _error, _warn
from debputy.version import debputy_doc_root_dir


def _provide_placeholder_parser_doc(
    parser_doc: ParserDocumentation | None,
    attributes: Iterable[str],
) -> ParserDocumentation:
    if parser_doc is None:
        parser_doc = reference_documentation()
    changes = {}
    if parser_doc.attribute_doc is None:
        changes["attribute_doc"] = [undocumented_attr(attr) for attr in attributes]

    if changes:
        return parser_doc.replace(**changes)
    return parser_doc


def doc_args_for_parser_doc(
    rule_name: str,
    declarative_parser: DeclarativeInputParser[Any],
    plugin_metadata: DebputyPluginMetadata,
    *,
    manifest_format_url: str | None = None,
) -> tuple[Mapping[str, str], ParserDocumentation]:
    attributes: Iterable[str]
    if isinstance(declarative_parser, DeclarativeMappingInputParser):
        attributes = declarative_parser.source_attributes.keys()
    else:
        attributes = []
    if manifest_format_url is None:
        manifest_format_url = f"{debputy_doc_root_dir()}/MANIFEST-FORMAT.md"
    doc_args = {
        "RULE_NAME": rule_name,
        "MANIFEST_FORMAT_DOC": manifest_format_url,
        "PLUGIN_NAME": plugin_metadata.plugin_name,
    }
    parser_doc = _provide_placeholder_parser_doc(
        declarative_parser.inline_reference_documentation,
        attributes,
    )
    return doc_args, parser_doc


def render_attribute_doc(
    parser: Any,
    attributes: Mapping[str, "AttributeDescription"],
    required_attributes: frozenset[str],
    conditionally_required_attributes: frozenset[frozenset[str]],
    parser_doc: ParserDocumentation,
    doc_args: Mapping[str, str],
    *,
    rule_name: str = "<unset>",
    is_root_rule: bool = False,
    is_interactive: bool = False,
) -> Iterable[tuple[frozenset[str], Sequence[str]]]:
    provided_attribute_docs = (
        parser_doc.attribute_doc if parser_doc.attribute_doc is not None else []
    )

    for attr_doc in assume_not_none(provided_attribute_docs):
        attr_description = attr_doc.description
        rendered_doc = []

        for parameter in sorted(attr_doc.attributes):
            parameter_details = attributes.get(parameter)
            if parameter_details is not None:
                source_name = parameter_details.source_attribute_name
                describe_type = parameter_details.type_validator.describe_type()
            else:
                assert isinstance(parser, DispatchingObjectParser)
                source_name = parameter
                subparser = parser.parser_for(source_name).parser
                if isinstance(subparser, InPackageContextParser):
                    if is_interactive:
                        describe_type = "PackageContext"
                    else:
                        rule_prefix = rule_name if not is_root_rule else ""
                        describe_type = f"PackageContext (chains to `{rule_prefix}::{subparser.manifest_attribute_path_template}`)"

                elif isinstance(subparser, DispatchingObjectParser):
                    if is_interactive:
                        describe_type = "Object"
                    else:
                        rule_prefix = rule_name if not is_root_rule else ""
                        describe_type = f"Object (see `{rule_prefix}::{subparser.manifest_attribute_path_template}`)"
                elif isinstance(subparser, DeclarativeMappingInputParser):
                    describe_type = "<Type definition not implemented yet>"  # TODO: Derive from subparser
                elif isinstance(subparser, DeclarativeNonMappingInputParser):
                    describe_type = (
                        subparser.alt_form_parser.type_validator.describe_type()
                    )
                else:
                    describe_type = f"<Unknown: Non-introspectable subparser - {subparser.__class__.__name__}>"

            if source_name in required_attributes:
                req_str = "required"
            elif any(source_name in s for s in conditionally_required_attributes):
                req_str = "conditional"
            else:
                req_str = "optional"
            rendered_doc.append(f"`{source_name}` ({req_str}): {describe_type}")

        if attr_description:
            rendered_doc.append("")
            attr_doc_rendered = _render_template(
                f"attr docs for {rule_name}",
                attr_description,
                doc_args,
            )
            rendered_doc.extend(
                line for line in attr_doc_rendered.splitlines(keepends=False)
            )
            rendered_doc.append("")
        yield attr_doc.attributes, rendered_doc


def _render_template(name: str, template_str: str, params: Mapping[str, str]) -> str:
    try:
        return Template(template_str).substitute(params)
    except KeyError as e:
        _warn(f"Render issue: {str(e)}")
        _error(f"Failed to render {name}: Missing key {e.args[0]}")
    except ValueError as e:
        _warn(f"Render issue: {str(e)}")
        _error(f"Failed to render {name}")


def _render_integration_mode(
    expected_modes: Container[DebputyIntegrationMode] | None,
) -> str | None:
    if expected_modes:
        allowed_modes = set()
        for mode in sorted(ALL_DEBPUTY_INTEGRATION_MODES):
            if mode in expected_modes:
                allowed_modes.add(mode)

        if allowed_modes == ALL_DEBPUTY_INTEGRATION_MODES:
            restriction = "any integration mode"
        else:
            restriction = ", ".join(sorted(allowed_modes))
    else:
        restriction = "any integration mode"

    return f"Integration mode availability: {restriction}"


def render_rule(
    rule_name: str,
    declarative_parser: DeclarativeInputParser[Any],
    plugin_metadata: DebputyPluginMetadata,
    color_base: OutputStyle,
    *,
    is_root_rule: bool = False,
    include_ref_doc_link: bool = True,
    include_alt_format: bool = True,
    base_heading_level: int = 1,
    manifest_format_url: str | None = None,
    show_integration_mode: bool = True,
) -> str:
    doc_args, parser_doc = doc_args_for_parser_doc(
        "the manifest root" if is_root_rule else rule_name,
        declarative_parser,
        plugin_metadata,
        manifest_format_url=manifest_format_url,
    )
    t = _render_template(
        f"title of {rule_name}",
        assume_not_none(parser_doc.title),
        doc_args,
    )
    body = _render_template(
        f"body of {rule_name}",
        assume_not_none(parser_doc.description),
        doc_args,
    ).rstrip()
    r = [
        color_base.heading(t, base_heading_level),
        "",
        body,
        "",
    ]

    if show_integration_mode:
        allowed_integration_modes = _render_integration_mode(
            declarative_parser.expected_debputy_integration_mode
        )
    else:
        allowed_integration_modes = None
    alt_form_parser = getattr(declarative_parser, "alt_form_parser", None)
    is_list_wrapped = False
    unwrapped_parser = declarative_parser
    if isinstance(declarative_parser, ListWrappedDeclarativeInputParser):
        is_list_wrapped = True
        unwrapped_parser = declarative_parser.delegate

    if isinstance(
        unwrapped_parser, (DeclarativeMappingInputParser, DispatchingObjectParser)
    ):

        if isinstance(unwrapped_parser, DeclarativeMappingInputParser):
            attributes = unwrapped_parser.source_attributes
            required = unwrapped_parser.input_time_required_parameters
            conditionally_required = unwrapped_parser.at_least_one_of
            mutually_exclusive = unwrapped_parser.mutually_exclusive_attributes
        else:
            attributes = {}
            required = frozenset()
            conditionally_required = frozenset()
            mutually_exclusive = frozenset()
        if is_list_wrapped:
            r.append("List where each element has the following attributes:")
        else:
            r.append("Attributes:")

        rendered_attr_doc = render_attribute_doc(
            unwrapped_parser,
            attributes,
            required,
            conditionally_required,
            parser_doc,
            doc_args,
            is_root_rule=is_root_rule,
            rule_name=rule_name,
            is_interactive=False,
        )
        for _, rendered_doc in rendered_attr_doc:
            prefix = " - "
            for line in rendered_doc:
                if line:
                    r.append(f"{prefix}{line}")
                else:
                    r.append("")
                prefix = "   "

        if (
            bool(conditionally_required)
            or bool(mutually_exclusive)
            or any(pd.conflicting_attributes for pd in attributes.values())
        ):
            r.append("")
            if is_list_wrapped:
                r.append(
                    "This rule enforces the following restrictions on each element in the list:"
                )
            else:
                r.append("This rule enforces the following restrictions:")

            if conditionally_required or mutually_exclusive:
                all_groups = list(
                    itertools.chain(conditionally_required, mutually_exclusive)
                )
                seen = set()
                for g in all_groups:
                    if g in seen:
                        continue
                    seen.add(g)
                    anames = "`, `".join(sorted(g))
                    is_mx = g in mutually_exclusive
                    is_cr = g in conditionally_required
                    if is_mx and is_cr:
                        r.append(f" - The rule must use exactly one of: `{anames}`")
                    elif is_cr:
                        r.append(f" - The rule must use at least one of: `{anames}`")
                    else:
                        assert is_mx
                        r.append(
                            f" - The following attributes are mutually exclusive: `{anames}`"
                        )

            if mutually_exclusive or any(
                pd.conflicting_attributes for pd in attributes.values()
            ):
                for parameter, parameter_details in sorted(attributes.items()):
                    source_name = parameter_details.source_attribute_name
                    conflicts = set(parameter_details.conflicting_attributes)
                    for mx in mutually_exclusive:
                        if parameter in mx and mx not in conditionally_required:
                            conflicts |= mx
                    if conflicts:
                        conflicts.discard(parameter)
                        cnames = "`, `".join(
                            sorted(
                                attributes[a].source_attribute_name for a in conflicts
                            )
                        )
                        r.append(
                            f" - The attribute `{source_name}` cannot be used with any of: `{cnames}`"
                        )
        r.append("")
    if include_alt_format and alt_form_parser is not None:
        # FIXME: Mapping[str, Any] ends here, which is ironic given the headline.
        r.append(
            f"Non-mapping format: {alt_form_parser.type_validator.describe_type()}"
        )
        alt_parser_desc = parser_doc.alt_parser_description
        if alt_parser_desc:
            r.extend(
                f"   {line}"
                for line in alt_parser_desc.format(**doc_args).splitlines(
                    keepends=False
                )
            )
        r.append("")

    if allowed_integration_modes:
        r.append(allowed_integration_modes)

    if include_ref_doc_link:
        if declarative_parser.reference_documentation_url is not None:
            r.append(
                f"Reference documentation: {declarative_parser.reference_documentation_url}"
            )
        else:
            r.append(
                "Reference documentation: No reference documentation link provided by the plugin"
            )
    elif allowed_integration_modes:
        # Better spacing in the generated docs, but it looks weird with this newline
        # in `debputy plugin show p-m-r ...`
        r.append("")

    return "\n".join(r)


def render_multiline_documentation(
    documentation: str,
    *,
    first_line_prefix: str = "Documentation: ",
    following_line_prefix: str = " ",
) -> Iterable[str]:
    current_prefix = first_line_prefix
    result = []
    for line in documentation.splitlines(keepends=False):
        if line.isspace():
            if not current_prefix.isspace():
                result.append(current_prefix.rstrip())
                current_prefix = following_line_prefix
            else:
                result.append("")
            continue
        result.append(f"{current_prefix}{line}")
        current_prefix = following_line_prefix
    return result


def _render_type_example(
    type_mapping: TypeMapping[Any, Any],
    output_style: OutputStyle,
    parser_context: ParserContextData,
    example: TypeMappingExample,
    *,
    recover_from_broken_examples: bool,
) -> tuple[str, bool]:
    attr_path = AttributePath.builtin_path()["Render Request"]
    v = _render_value(example.source_input)
    try:
        type_mapping.mapper(
            example.source_input,
            attr_path,
            parser_context,
        )
    except RuntimeError:
        if not recover_from_broken_examples:
            raise
        return (
            output_style.colored(v, fg="red") + " [Example value could not be parsed]",
            True,
        )
    return output_style.colored(v, fg="green"), False


def render_source_type(t: Any) -> str:
    _, origin_type, args = unpack_type(t, False)
    if origin_type == Union:
        return " | ".join(render_source_type(st) for st in args)
    name = BASIC_SIMPLE_TYPES.get(t)
    if name is not None:
        return name
    try:
        return t.__name__
    except AttributeError:
        return str(t)


def render_type_mapping(
    pptm: PluginProvidedTypeMapping,
    output_style: OutputStyle,
    parser_context: ParserContextData,
    *,
    recover_from_broken_examples: bool = False,
    base_heading_level: int = 1,
) -> str:
    type_mapping = pptm.mapped_type
    target_type = type_mapping.target_type
    ref_doc = pptm.reference_documentation
    desc = ref_doc.description if ref_doc is not None else None
    examples = ref_doc.examples if ref_doc is not None else tuple()
    base_type = render_source_type(type_mapping.source_type)
    lines = [
        output_style.heading(
            f"Type Mapping: {target_type.__name__} [{base_type}]", base_heading_level
        ),
        "",
    ]

    if desc is not None:
        lines.extend(
            render_multiline_documentation(
                desc,
                first_line_prefix="",
                following_line_prefix="",
            )
        )
    else:
        lines.append("No documentation provided.")

    if examples:
        had_issues = False
        lines.append("")
        lines.append(output_style.heading("Example values", base_heading_level + 1))
        lines.append("")
        for no, example in enumerate(examples, start=1):
            v, i = _render_type_example(
                type_mapping,
                output_style,
                parser_context,
                example,
                recover_from_broken_examples=recover_from_broken_examples,
            )
            if i and recover_from_broken_examples:
                lines.append(
                    output_style.colored("Broken example: ", fg="red")
                    + f"Provided example input ({v})"
                    + " caused an exception when parsed. Please file a bug against the plugin."
                    + " Use --debug/DEBPUTY_DEBUG=1 to see the stack trace"
                )
            lines.append(f" * {v}")
            if i:
                had_issues = True
    else:
        had_issues = False

    if had_issues:
        lines.append("")
        lines.append(
            output_style.colored(
                "Examples had issues. Please file a bug against the plugin", fg="red"
            )
        )
        lines.append("")
        lines.append("Use --debug/DEBPUTY_DEBUG=1 to see the stacktrace")

    return "\n".join(lines)


def _render_value(v: Any) -> str:
    if isinstance(v, str) and '"' not in v:
        return f'"{v}"'
    return str(v)



import dataclasses
import json
import os
import stat
import sys
import tarfile
from enum import Enum


from typing import Optional, List, Dict, Any, Union, Self, IO
from collections.abc import Iterable, Mapping

IntermediateManifest = list["TarMember"]


class PathType(Enum):
    FILE = ("file", tarfile.REGTYPE)
    DIRECTORY = ("directory", tarfile.DIRTYPE)
    SYMLINK = ("symlink", tarfile.SYMTYPE)
    # TODO: Add hardlink, FIFO, Char device, BLK device, etc.

    @property
    def manifest_key(self) -> str:
        return self.value[0]

    @property
    def tarinfo_type(self) -> bytes:
        return self.value[1]

    @property
    def can_be_virtual(self) -> bool:
        return self in (PathType.DIRECTORY, PathType.SYMLINK)


KEY2PATH_TYPE = {pt.manifest_key: pt for pt in PathType}


def _dirname(path: str) -> str:
    path = path.rstrip("/")
    if path == ".":
        return path
    return os.path.dirname(path)


def _fs_type_from_st_mode(fs_path: str, st_mode: int) -> PathType:
    if stat.S_ISREG(st_mode):
        path_type = PathType.FILE
    elif stat.S_ISDIR(st_mode):
        path_type = PathType.DIRECTORY
    #        elif stat.S_ISFIFO(st_result):
    #            type = FIFOTYPE
    elif stat.S_ISLNK(st_mode):
        raise ValueError(
            "Symlinks should have been rewritten to use the virtual rule."
            " Otherwise, the link would not be normalized according to Debian Policy."
        )
    #        elif stat.S_ISCHR(st_result):
    #            type = CHRTYPE
    #        elif stat.S_ISBLK(st_result):
    #            type = BLKTYPE
    else:
        raise ValueError(
            f"The path {fs_path} had an unsupported/unknown file type."
            f" Probably a bug in the tool"
        )
    return path_type


@dataclasses.dataclass(slots=True)
class TarMember:
    member_path: str
    path_type: PathType
    fs_path: str | None
    mode: int
    owner: str
    uid: int
    group: str
    gid: int
    mtime: float
    link_target: str = ""
    is_virtual_entry: bool = False
    may_steal_fs_path: bool = False

    def create_tar_info(self, tar_fd: tarfile.TarFile) -> tarfile.TarInfo:
        tar_info: tarfile.TarInfo
        if self.is_virtual_entry:
            assert self.path_type.can_be_virtual
            tar_info = tar_fd.tarinfo(self.member_path)
            tar_info.size = 0
            tar_info.type = self.path_type.tarinfo_type
            tar_info.linkpath = self.link_target
        else:
            try:
                tar_info = tar_fd.gettarinfo(
                    name=self.fs_path, arcname=self.member_path
                )
            except (TypeError, ValueError) as e:
                raise ValueError(
                    f"Unable to prepare tar info for {self.member_path}"
                ) from e
            # TODO: Eventually, we should be able to unconditionally rely on link_target.  However,
            # until we got symlinks and hardlinks correctly done in the JSON generator, it will be
            # conditional for now.
            if self.link_target != "":
                tar_info.linkpath = self.link_target
        tar_info.mode = self.mode
        tar_info.uname = self.owner
        tar_info.uid = self.uid
        tar_info.gname = self.group
        tar_info.gid = self.gid
        tar_info.mode = self.mode
        tar_info.mtime = int(self.mtime)

        return tar_info

    @classmethod
    def from_file(
        cls,
        member_path: str,
        fs_path: str,
        mode: int | None = None,
        owner: str = "root",
        uid: int = 0,
        group: str = "root",
        gid: int = 0,
        path_mtime: float | int | None = None,
        clamp_mtime_to: int | None = None,
        path_type: PathType | None = None,
        may_steal_fs_path: bool = False,
    ) -> "TarMember":
        # Avoid lstat'ing if we can as it makes it easier to do tests of the code
        # (as we do not need an existing physical fs path)
        if path_type is None or path_mtime is None or mode is None:
            st_result = os.lstat(fs_path)
            st_mode = st_result.st_mode
            if mode is None:
                mode = st_mode
            if path_mtime is None:
                path_mtime = st_result.st_mtime
            if path_type is None:
                path_type = _fs_type_from_st_mode(fs_path, st_mode)

        if clamp_mtime_to is not None and path_mtime > clamp_mtime_to:
            path_mtime = clamp_mtime_to

        if may_steal_fs_path:
            assert (
                "debputy/scratch-dir/" in fs_path
            ), f"{fs_path} should not have been stealable"

        return cls(
            member_path=member_path,
            path_type=path_type,
            fs_path=fs_path,
            mode=mode,
            owner=owner,
            uid=uid,
            group=group,
            gid=gid,
            mtime=float(path_mtime),
            is_virtual_entry=False,
            may_steal_fs_path=may_steal_fs_path,
        )

    @classmethod
    def virtual_path(
        cls,
        member_path: str,
        path_type: PathType,
        mtime: float,
        mode: int,
        link_target: str = "",
        owner: str = "root",
        uid: int = 0,
        group: str = "root",
        gid: int = 0,
    ) -> Self:
        if not path_type.can_be_virtual:
            raise ValueError(f"The path type {path_type.name} cannot be virtual")
        if (path_type == PathType.SYMLINK) ^ bool(link_target):
            if not link_target:
                raise ValueError("Symlinks must have a link target")
            # TODO: Dear future programmer. Hardlinks will appear here some day and you will have to fix this
            # code then!
            raise ValueError("Non-symlinks must not have a link target")
        return cls(
            member_path=member_path,
            path_type=path_type,
            fs_path=None,
            link_target=link_target,
            mode=mode,
            owner=owner,
            uid=uid,
            group=group,
            gid=gid,
            mtime=mtime,
            is_virtual_entry=True,
        )

    def clone_and_replace(self, /, **changes: Any) -> "TarMember":
        return dataclasses.replace(self, **changes)

    def to_manifest(self) -> dict[str, Any]:
        d = dataclasses.asdict(self)
        try:
            d["mode"] = oct(self.mode)
        except (TypeError, ValueError) as e:
            raise TypeError(f"Bad mode in TarMember {self.member_path}") from e
        d["path_type"] = self.path_type.manifest_key
        # "compress" the output by removing redundant fields
        if self.link_target is None or self.link_target == "":
            del d["link_target"]
        if self.is_virtual_entry:
            assert self.fs_path is None
            del d["fs_path"]
        else:
            del d["is_virtual_entry"]
        return d

    @classmethod
    def parse_intermediate_manifest(cls, manifest_path: str) -> IntermediateManifest:
        directories = {"."}
        if manifest_path == "-":
            with sys.stdin as fd:
                data = json.load(fd)
                contents = [TarMember.from_dict(m) for m in data]
        else:
            with open(manifest_path) as fd:
                data = json.load(fd)
                contents = [TarMember.from_dict(m) for m in data]
        if not contents:
            raise ValueError(
                "Empty manifest (note that the root directory should always be present"
            )
        if contents[0].member_path != "./":
            raise ValueError('The first member must always be the root directory "./"')
        for tar_member in contents:
            directory = _dirname(tar_member.member_path)
            if directory not in directories:
                raise ValueError(
                    f'The path "{tar_member.member_path}" came before the directory it is in (or the path'
                    f" is not a directory). Either way leads to a broken deb."
                )
            if tar_member.path_type == PathType.DIRECTORY:
                directories.add(tar_member.member_path.rstrip("/"))
        return contents

    @classmethod
    def from_dict(cls, d: Any) -> "TarMember":
        member_path = d["member_path"]
        raw_mode = d["mode"]
        if not raw_mode.startswith("0o"):
            raise ValueError(f"Bad mode for {member_path}")
        is_virtual_entry = d.get("is_virtual_entry") or False
        path_type = KEY2PATH_TYPE[d["path_type"]]
        fs_path = d.get("fs_path")
        mode = int(raw_mode[2:], 8)
        if is_virtual_entry:
            if not path_type.can_be_virtual:
                raise ValueError(
                    f"Bad file type or is_virtual_entry for {d['member_path']}."
                    " The file type cannot be virtual"
                )
            if fs_path is not None:
                raise ValueError(
                    f'Invalid declaration for "{member_path}".'
                    " The path is listed as a virtual entry but has a file system path"
                )
        elif fs_path is None:
            raise ValueError(
                f'Invalid declaration for "{member_path}".'
                " The path is neither a virtual path nor does it have a file system path!"
            )
        if path_type == PathType.DIRECTORY and not member_path.endswith("/"):
            raise ValueError(
                f'Invalid declaration for "{member_path}".'
                " The path is listed as a directory but does not end with a slash"
            )

        link_target = d.get("link_target")
        if path_type == PathType.SYMLINK:
            if mode != 0o777:
                raise ValueError(
                    f'Invalid declaration for "{member_path}".'
                    f" Symlinks must have mode 0o0777, got {oct(mode)[2:]}."
                )
            if not link_target:
                raise ValueError(
                    f'Invalid declaration for "{member_path}".'
                    " Symlinks must have a link_target"
                )
        elif link_target is not None and link_target != "":
            # TODO: Eventually hardlinks should have them too.  But that is a problem for a future programmer
            raise ValueError(
                f'Invalid declaration for "{member_path}".'
                " Only symlinks can have a link_target"
            )
        else:
            link_target = ""
        may_steal_fs_path = d.get("may_steal_fs_path") or False

        if may_steal_fs_path:
            assert (
                "debputy/scratch-dir/" in fs_path
            ), f"{fs_path} should not have been stealable"
        return cls(
            member_path=member_path,
            path_type=path_type,
            fs_path=fs_path,
            mode=mode,
            owner=d["owner"],
            uid=d["uid"],
            group=d["group"],
            gid=d["gid"],
            mtime=float(d["mtime"]),
            link_target=link_target,
            is_virtual_entry=is_virtual_entry,
            may_steal_fs_path=may_steal_fs_path,
        )


def output_intermediate_manifest(
    manifest_output_file: str,
    members: Iterable[TarMember],
) -> None:
    with open(manifest_output_file, "w") as fd:
        output_intermediate_manifest_to_fd(fd, members)


def output_intermediate_manifest_to_fd(
    fd: IO[str], members: Iterable[TarMember]
) -> None:
    serial_format = [m.to_manifest() for m in members]
    json.dump(serial_format, fd)



import os.path
import subprocess

from debian.debian_support import Version

from debputy.analysis.debian_dir import resolve_debhelper_config_files
from debputy.dh.debhelper_emulation import dhe_pkgfile
from debputy.dh.dh_assistant import read_dh_addon_sequences, extract_dh_compat_level
from debputy.dh_migration.models import (
    FeatureMigration,
    MigrationRequest,
)
from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin


def _installed_debhelper_version_is_at_least(desired_version: str) -> bool:
    try:
        output = subprocess.check_output(
            [
                "dpkg-query",
                "-W",
                "--showformat=${Version} ${db:Status-Status}\n",
                "debhelper",
            ]
        ).decode("utf-8")
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False
    else:
        parts = output.split()
        if len(parts) != 2:
            return False
        if parts[1] != "installed":
            return False
        return Version(parts[0]) >= Version(desired_version)


def dh_mbm_migrate_package_files(
    migration_request: MigrationRequest,
    feature_migration: FeatureMigration,
) -> None:
    feature_migration.tagline = "Add explicit package name to debhelper config files"
    debputy_plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    dh_compat_level, _ = extract_dh_compat_level()
    r = read_dh_addon_sequences(migration_request.debian_dir)
    if r is not None:
        bd_sequences, dr_sequences, uses_dh_sequencer = r
        dh_sequences = bd_sequences | dr_sequences
        uses_dh_sequencer = True
    else:
        dh_sequences = set()
        uses_dh_sequencer = False

    (
        all_dh_ppfs,
        missing_introspection,
        _,
        _,
    ) = resolve_debhelper_config_files(
        migration_request.debian_dir,
        {p.name: p for p in migration_request.all_packages},
        debputy_plugin_metadata,
        migration_request.manifest.plugin_provided_feature_set,
        dh_sequences,
        dh_compat_level,
        uses_dh_sequencer,
    )
    for ppf in all_dh_ppfs:
        if (
            ppf.uses_explicit_package_name
            or ppf.definition.packageless_is_fallback_for_all_packages
        ):
            continue
        basename = ppf.path.name
        new_basename = f"{ppf.package_name}.{basename}"
        ppf_parent = ppf.path.parent_dir
        assert ppf_parent is not None
        parent_path = ppf_parent.path
        new_path = os.path.join(parent_path, new_basename)
        feature_migration.rename_on_success(ppf.path.path, new_path)

    if not missing_introspection and not _installed_debhelper_version_is_at_least(
        "13.26~"
    ):
        feature_migration.warn(
            "Please upgrade to `debhelper/13.26` to ensure reliability. Some files might be silently missed."
        )

    for command in missing_introspection:
        feature_migration.warn(
            f"The command `{command}` is used but `dh_assistant` was not aware of whether it had configuration files."
        )

    if missing_introspection:
        feature_migration.warn(
            "Please ensure you have the latest versions of the above commands installed and try again."
        )
        feature_migration.warn(
            "If the issue still persists, then please file a bug against the provider of the commands asking them to supply the relevant hints (see doc/PROGRAMMING.md in debhelper/13.26+) with the debhelper maintainers in CC."
        )
    feature_migration.warn(
        "Any architecture restricted file (such `debian/install.amd64`) and templating based files (such as debian/install.tmpl) may be missed. Please validate these manually."
    )


def dh_mbm_migrate_manual_warnings(
    migration_request: MigrationRequest,
    feature_migration: FeatureMigration,
) -> None:
    feature_migration.tagline = "Manual parts"
    debian_dir = migration_request.debian_dir
    r = read_dh_addon_sequences(debian_dir)
    if r is not None:
        bd_sequences, dr_sequences, uses_dh_sequencer = r
        dh_sequences = bd_sequences | dr_sequences
        uses_dh_sequencer = True
    else:
        dh_sequences = set()
        uses_dh_sequencer = False
    drules = debian_dir.get("rules")
    if drules and drules.is_file:
        with drules.open() as fd:
            migrated_auto_install = 0
            for line in fd:
                line = line.lstrip()
                if not line or line.startswith("#"):
                    continue
                if line.startswith("\t") and "dh_auto_install" in line:
                    migrated_auto_install += 1 if "--destdir" in line else 0
    used_dh_install_files = 0
    for dctrl_bin in migration_request.all_packages:
        dh_config_file = dhe_pkgfile(debian_dir, dctrl_bin, "install")
        if dh_config_file is not None:
            used_dh_install_files += 1
    main_pkg = migration_request.main_binary
    if not migrated_auto_install and not used_dh_install_files:
        feature_migration.warn(
            f"Ensure `dh_auto_install --destdir debian/{main_pkg.name}` or `dh_install` is used. Remember Replaces/Breaks if files are moved"
        )
    if uses_dh_sequencer and "single-binary" in dh_sequences:
        feature_migration.warn("Remove `single-binary` add-on")



from typing import (
    Literal,
    TypedDict,
    Union,
    TypeVar,
    Dict,
    Optional,
    List,
    cast,
    NotRequired,
    TYPE_CHECKING,
)
from collections.abc import Callable, Iterable, Mapping

from debputy.lsp.diagnostics import DiagnosticData
from debputy.util import _warn

if TYPE_CHECKING:
    import lsprotocol.types as types
    from debputy.linting.lint_util import LintState
else:
    import debputy.lsprotocol.types as types


try:
    from debian._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )

    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument
    from debputy.lsp.debputy_ls import DebputyLanguageServer
except ImportError:
    pass


CodeActionName = Literal[
    "correct-text",
    "remove-line",
    "remove-range",
    "insert-text-on-line-after-diagnostic",
]


class CorrectTextCodeAction(TypedDict):
    code_action: Literal["correct-text"]
    correct_value: str
    proposed_title: NotRequired[str]


class InsertTextOnLineAfterDiagnosticCodeAction(TypedDict):
    code_action: Literal["insert-text-on-line-after-diagnostic"]
    text_to_insert: str


class RemoveLineCodeAction(TypedDict):
    code_action: Literal["remove-line"]


class RemoveRangeCodeAction(TypedDict):
    code_action: Literal["remove-range"]
    proposed_title: NotRequired[str]


def propose_correct_text_quick_fix(
    correct_value: str,
    *,
    proposed_title: str | None = None,
) -> CorrectTextCodeAction:
    r: CorrectTextCodeAction = {
        "code_action": "correct-text",
        "correct_value": correct_value,
    }
    if proposed_title:
        r["proposed_title"] = proposed_title
    return r


def propose_insert_text_on_line_after_diagnostic_quick_fix(
    text_to_insert: str,
) -> InsertTextOnLineAfterDiagnosticCodeAction:
    return {
        "code_action": "insert-text-on-line-after-diagnostic",
        "text_to_insert": text_to_insert,
    }


def propose_remove_line_quick_fix() -> RemoveLineCodeAction:
    return {
        "code_action": "remove-line",
    }


def propose_remove_range_quick_fix(
    *,
    proposed_title: str | None = None,
) -> RemoveRangeCodeAction:
    r: RemoveRangeCodeAction = {
        "code_action": "remove-range",
    }
    if proposed_title:
        r["proposed_title"] = proposed_title
    return r


CODE_ACTION_HANDLERS: dict[
    CodeActionName,
    Callable[
        ["LintState", Mapping[str, str], types.CodeActionParams, types.Diagnostic],
        Iterable[types.CodeAction | types.Command],
    ],
] = {}
M = TypeVar("M", bound=Mapping[str, str])
Handler = Callable[
    ["LintState", M, types.CodeActionParams, types.Diagnostic],
    Iterable[Union[types.CodeAction, types.Command]],
]


def _code_handler_for(action_name: CodeActionName) -> Callable[[Handler], Handler]:
    def _wrapper(func: Handler) -> Handler:
        assert action_name not in CODE_ACTION_HANDLERS
        CODE_ACTION_HANDLERS[action_name] = func
        return func

    return _wrapper


@_code_handler_for("correct-text")
def _correct_value_code_action(
    lint_state: "LintState",
    code_action_data: CorrectTextCodeAction,
    code_action_params: types.CodeActionParams,
    diagnostic: types.Diagnostic,
) -> Iterable[types.CodeAction | types.Command]:
    corrected_value = code_action_data["correct_value"]
    title = code_action_data.get("proposed_title")
    if not title:
        title = f'Replace with "{corrected_value}"'
    yield _simple_quick_fix(
        lint_state,
        code_action_params,
        title,
        [diagnostic],
        diagnostic.range,
        corrected_value,
    )


@_code_handler_for("insert-text-on-line-after-diagnostic")
def _insert_text_on_line_after_diagnostic_code_action(
    lint_state: "LintState",
    code_action_data: InsertTextOnLineAfterDiagnosticCodeAction,
    code_action_params: types.CodeActionParams,
    diagnostic: types.Diagnostic,
) -> Iterable[types.CodeAction | types.Command]:
    corrected_value = code_action_data["text_to_insert"]
    line_no = diagnostic.range.end.line
    if diagnostic.range.end.character > 0:
        line_no += 1
    insert_range = types.Range(
        types.Position(
            line_no,
            0,
        ),
        types.Position(
            line_no,
            0,
        ),
    )
    yield _simple_quick_fix(
        lint_state,
        code_action_params,
        f'Insert "{corrected_value}"',
        [diagnostic],
        insert_range,
        corrected_value,
    )


def _simple_quick_fix(
    lint_state: "LintState",
    _code_action_params: types.CodeActionParams,
    title: str,
    diagnostics: list[types.Diagnostic],
    affected_range: types.Range,
    replacement_text: str,
) -> types.CodeAction:
    doc_uri = lint_state.doc_uri
    edit = types.TextEdit(
        affected_range,
        replacement_text,
    )
    if lint_state.workspace_text_edit_support.supports_document_changes:
        ws_edit = types.WorkspaceEdit(
            document_changes=[
                types.TextDocumentEdit(
                    text_document=types.OptionalVersionedTextDocumentIdentifier(
                        doc_uri,
                        lint_state.doc_version,
                    ),
                    edits=[edit],
                )
            ]
        )
    else:
        ws_edit = types.WorkspaceEdit(
            changes={doc_uri: [edit]},
        )
    return types.CodeAction(
        title=title,
        kind=types.CodeActionKind.QuickFix,
        diagnostics=diagnostics,
        edit=ws_edit,
    )


def range_compatible_with_remove_line_fix(range_: types.Range | TERange) -> bool:
    if isinstance(range_, TERange):
        start = range_.start_pos
        end = range_.end_pos
        if start.line_position != end.line_position and (
            start.line_position + 1 != end.line_position or end.cursor_position > 0
        ):
            return False
    else:
        start = range_.start
        end = range_.end
        if start.line != end.line and (start.line + 1 != end.line or end.character > 0):
            return False
    return True


@_code_handler_for("remove-line")
def _remove_line_code_action(
    lint_state: "LintState",
    _code_action_data: RemoveLineCodeAction,
    code_action_params: types.CodeActionParams,
    diagnostic: types.Diagnostic,
) -> Iterable[types.CodeAction | types.Command]:
    start = code_action_params.range.start
    if range_compatible_with_remove_line_fix(code_action_params.range):
        _warn(
            "Bug: the quick was used for a diagnostic that spanned multiple lines and would corrupt the file."
        )
        return

    delete_range = types.Range(
        start=types.Position(
            line=start.line,
            character=0,
        ),
        end=types.Position(
            line=start.line + 1,
            character=0,
        ),
    )
    yield _simple_quick_fix(
        lint_state,
        code_action_params,
        "Remove the line",
        [diagnostic],
        delete_range,
        "",
    )


@_code_handler_for("remove-range")
def _remove_range_code_action(
    lint_state: "LintState",
    code_action_data: RemoveRangeCodeAction,
    code_action_params: types.CodeActionParams,
    diagnostic: types.Diagnostic,
) -> Iterable[types.CodeAction | types.Command]:
    title = code_action_data.get("proposed_title", "Delete")
    yield _simple_quick_fix(
        lint_state,
        code_action_params,
        title,
        [diagnostic],
        diagnostic.range,
        "",
    )


def accepts_quickfixes(
    code_action_params: types.CodeActionParams,
) -> bool:
    only = code_action_params.context.only
    if not only:
        return True
    return types.CodeActionKind.QuickFix in only


def provide_standard_quickfixes_from_diagnostics_ls(
    ls: "DebputyLanguageServer",
    code_action_params: types.CodeActionParams,
) -> list[types.Command | types.CodeAction] | None:
    if not accepts_quickfixes(code_action_params):
        return None
    doc_uri = code_action_params.text_document.uri
    matched_diagnostics = ls.diagnostics_in_range(
        doc_uri,
        code_action_params.range,
    )
    if not matched_diagnostics:
        return None
    doc = ls.workspace.get_text_document(doc_uri)
    return _provide_standard_quickfixes_from_diagnostics(
        ls.lint_state(doc),
        code_action_params,
        matched_diagnostics,
    )


def provide_standard_quickfixes_from_diagnostics_lint(
    lint_state: "LintState",
    code_action_params: types.CodeActionParams,
) -> list[types.Command | types.CodeAction] | None:
    return _provide_standard_quickfixes_from_diagnostics(
        lint_state,
        code_action_params,
        code_action_params.context.diagnostics,
    )


def _provide_standard_quickfixes_from_diagnostics(
    lint_state: "LintState",
    code_action_params: types.CodeActionParams,
    matched_diagnostics: list[types.Diagnostic],
) -> list[types.Command | types.CodeAction] | None:
    actions: list[types.Command | types.CodeAction] = []
    for diagnostic in matched_diagnostics:
        if not isinstance(diagnostic.data, dict):
            continue
        data: DiagnosticData = cast("DiagnosticData", diagnostic.data)
        quickfixes = data.get("quickfixes")
        if quickfixes is None:
            continue
        for action_suggestion in quickfixes:
            if (
                action_suggestion
                and isinstance(action_suggestion, Mapping)
                and "code_action" in action_suggestion
            ):
                action_name: CodeActionName = action_suggestion["code_action"]
                handler = CODE_ACTION_HANDLERS.get(action_name)
                if handler is not None:
                    actions.extend(
                        handler(
                            lint_state,
                            cast("Mapping[str, str]", action_suggestion),
                            code_action_params,
                            diagnostic,
                        )
                    )
                else:
                    _warn(f"No codeAction handler for {action_name} !?")
    if not actions:
        return None
    return actions



import collections.abc
import contextlib
import contextvars
import functools
import inspect
from collections.abc import Callable
from contextvars import ContextVar
from typing import ParamSpec, TypeVar, NoReturn, Any

from debputy.exceptions import (
    UnhandledOrUnexpectedErrorFromPluginError,
    DebputyRuntimeError,
)
from debputy.packages import SourcePackage, BinaryPackage
from debputy.util import _trace_log, _is_trace_log_enabled

_current_debputy_plugin_cxt_var: ContextVar[str | None] = ContextVar(
    "current_debputy_plugin",
    default=None,
)
_current_debputy_parsing_context: ContextVar[
    tuple[
        dict[tuple[SourcePackage | BinaryPackage, type[Any]], Any],
        SourcePackage | BinaryPackage,
    ]
    | None
] = ContextVar(
    "current_debputy_parsing_context",
    default=None,
)

P = ParamSpec("P")
R = TypeVar("R")


def register_manifest_type_value_in_context(
    value_type: type[Any],
    value: Any,
) -> None:
    context_vars = _current_debputy_parsing_context.get()
    if context_vars is None:
        raise AssertionError(
            "register_manifest_type_value_in_context() was called, but no context was set."
        )
    value_table, context_pkg = context_vars
    if (context_pkg, value_type) in value_table:
        raise AssertionError(
            f"The type {value_type!r} was already registered for {context_pkg}, which the plugin API should have prevented"
        )
    value_table[(context_pkg, value_type)] = value


def begin_parsing_context(
    value_table: dict[tuple[SourcePackage | BinaryPackage, type[Any]], Any],
    context_pkg: SourcePackage,
    func: Callable[P, R],
    *args: P.args,
    **kwargs: P.kwargs,
) -> R:
    context = contextvars.copy_context()
    # Wish we could just do a regular set without wrapping it in `context.run`
    context.run(_current_debputy_parsing_context.set, (value_table, context_pkg))
    assert context.get(_current_debputy_parsing_context) == (value_table, context_pkg)
    return context.run(func, *args, **kwargs)


@contextlib.contextmanager
def with_binary_pkg_parsing_context(
    context_pkg: BinaryPackage,
) -> collections.abc.Iterator[None]:
    context_vars = _current_debputy_parsing_context.get()
    if context_vars is None:
        raise AssertionError(
            "with_binary_pkg_parsing_context() was called, but no context was set."
        )
    value_table, _ = context_vars
    token = _current_debputy_parsing_context.set((value_table, context_pkg))
    try:
        yield
    finally:
        _current_debputy_parsing_context.reset(token)


def current_debputy_plugin_if_present() -> str | None:
    return _current_debputy_plugin_cxt_var.get()


def current_debputy_plugin_required() -> str:
    v = current_debputy_plugin_if_present()
    if v is None:
        raise AssertionError(
            "current_debputy_plugin_required() was called, but no plugin was set."
        )
    return v


def wrap_plugin_code(
    plugin_name: str,
    func: Callable[P, R],
    *,
    non_debputy_exception_handling: bool | Callable[[Exception], NoReturn] = True,
) -> Callable[P, R]:
    if isinstance(non_debputy_exception_handling, bool):

        runner = run_in_context_of_plugin
        if non_debputy_exception_handling:
            runner = run_in_context_of_plugin_wrap_errors

        def _plugin_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return runner(plugin_name, func, *args, **kwargs)

        functools.update_wrapper(_plugin_wrapper, func)
        return _plugin_wrapper

    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        try:
            return run_in_context_of_plugin(plugin_name, func, *args, **kwargs)
        except DebputyRuntimeError:
            raise
        except Exception as e:
            non_debputy_exception_handling(e)

    functools.update_wrapper(_wrapper, func)
    return _wrapper


def run_in_context_of_plugin(
    plugin: str,
    func: Callable[P, R],
    *args: P.args,
    **kwargs: P.kwargs,
) -> R:
    context = contextvars.copy_context()
    if _is_trace_log_enabled():
        call_stack = inspect.stack()
        caller: str = "[N/A]"
        for frame in call_stack:
            if frame.filename != __file__:
                try:
                    fname = frame.frame.f_code.co_qualname
                except AttributeError:
                    fname = None
                if fname is None:
                    fname = frame.function
                caller = f"{frame.filename}:{frame.lineno} ({fname})"
                break
        # Do not keep the reference longer than necessary
        del call_stack
        _trace_log(
            f"Switching plugin context to {plugin} at {caller} (from context: {current_debputy_plugin_if_present()})"
        )
    # Wish we could just do a regular set without wrapping it in `context.run`
    context.run(_current_debputy_plugin_cxt_var.set, plugin)
    return context.run(func, *args, **kwargs)


def run_in_context_of_plugin_wrap_errors(
    plugin: str,
    func: Callable[P, R],
    *args: P.args,
    **kwargs: P.kwargs,
) -> R:
    try:
        return run_in_context_of_plugin(plugin, func, *args, **kwargs)
    except DebputyRuntimeError:
        raise
    except Exception as e:
        if plugin != "debputy":
            raise UnhandledOrUnexpectedErrorFromPluginError(
                f"{func.__qualname__} from the plugin {plugin} raised exception that was not expected here."
            ) from e
        else:
            raise AssertionError(
                "Bug in the `debputy` plugin: Unhandled exception."
            ) from e



from debputy.maintscript_snippet import (
    MaintscriptSnippetContainer,
    MaintscriptSnippet,
    SnippetResolver,
)


def test_snippet_generation() -> None:
    container = MaintscriptSnippetContainer()

    assert container.generate_snippet() is None
    assert container.generate_snippet(snippet_order="service") is None

    container.append(
        MaintscriptSnippet(
            definition_source="test/1",
            snippet=SnippetResolver.snippet_template("systemctl status\n"),
            snippet_order="service",
        )
    )
    container.append(
        MaintscriptSnippet(
            definition_source="test/2",
            snippet=SnippetResolver.snippet_template("run foo\n"),
        )
    )
    assert not container.needs_debconf()
    snippet = container.generate_snippet()
    assert snippet
    assert "run foo\n" in snippet
    assert "systemctl status\n" not in snippet
    service_snippet = container.generate_snippet(snippet_order="service")
    assert service_snippet
    assert "run foo\n" not in service_snippet
    assert "systemctl status\n" in service_snippet

    container.append(
        MaintscriptSnippet(
            definition_source="test/2",
            snippet=SnippetResolver.snippet_template("db_get foo\n"),
            uses_debconf=True,
        )
    )

    assert container.needs_debconf()
    snippet = container.generate_snippet()
    assert snippet
    assert "run foo\n" in snippet
    assert "db_get foo\n" in snippet
    assert "systemctl status\n" not in snippet
    service_snippet = container.generate_snippet(snippet_order="service")
    assert service_snippet
    assert "run foo\n" not in service_snippet
    assert "db_get foo\n" not in service_snippet
    assert "systemctl status\n" in service_snippet
    snippet_lines = snippet.splitlines(keepends=True)
    foo_idx = snippet_lines.index("run foo\n")
    db_get_idx = snippet_lines.index("db_get foo\n")
    assert foo_idx < db_get_idx

    reversed_snippet = container.generate_snippet(reverse=True)
    assert reversed_snippet
    assert "run foo\n" in reversed_snippet
    assert "db_get foo\n" in reversed_snippet
    assert "systemctl status\n" not in reversed_snippet

    reversed_snippet_lines = reversed_snippet.splitlines(keepends=True)
    reversed_foo_idx = reversed_snippet_lines.index("run foo\n")
    reversed_db_get_idx = reversed_snippet_lines.index("db_get foo\n")
    assert reversed_foo_idx > reversed_db_get_idx
